"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import torch
import torch.nn.functional as F
from ...common.registry import registry
from ..base_model import tile
from .blip import BlipBase
from .blip_outputs import (
    BlipOutput,
    BlipIntermediateOutput,
)
from ..med import XBertEncoder, XBertLMHeadDecoder
from ..vit import VisionTransformerEncoder


@registry.register_model("blip_vqa")
class BlipVQA(BlipBase):
    """
    BLIP VQA models.

    Supported model types:
        - base: vqa model initialized with pre-trained BLIP base model on 115M image-text pairs after CapFilt; not fine-tuned.
        - vqav2: fine-tuned BLIP base model on VQA v2.0 dataset.

    Usage:
        >>> from lavis.models import load_model
        >>> model = load_model("blip_vqa", "vqav2")
        >>> model = load_model("blip_vqa", "okvqa")
        >>> model = load_model("blip_vqa", "aokvqa")
    """

    PRETRAINED_MODEL_CONFIG_DICT = {
        "vqav2": "configs/models/blip_vqav2.yaml",
        "okvqa": "configs/models/blip_vqa_okvqa.yaml",
        "aokvqa": "configs/models/blip_vqa_aokvqa.yaml",
    }

    def __init__(self, image_encoder, text_encoder, text_decoder, max_txt_len=35):
        super().__init__()
        self.tokenizer = self.init_tokenizer()

        self.visual_encoder = image_encoder

        self.text_encoder = text_encoder
        self.text_decoder = text_decoder

        self.max_txt_len = max_txt_len

    def forward(self, samples):
        """
        Args:
            samples (dict): A dictionary containing the following keys:
                - image (torch.Tensor): A tensor of shape (batch_size, 3, H, W). Default H=480, W=480.
                - text_input (list): A list of strings, each string is a question
                - answer (list): A list of strings, each string is an answer
                - weight (torch.Tensor): A tensor used to weigh each answer in the loss computation.
                   The shape of the tensor is (sum(n_answers),)
                - n_answers (torch.Tensor): A tensor shape (batch_size,) containing the number of answers
                     for each question in the batch.

        Returns:
            A BlipOutput object containing loss and intermediate outputs,
            see :class:`lavis.models.blip_outputs.BlipOutput` for more details.

        Examples:
        ```python
            >>> import torch
            >>> from lavis.models import load_model
            >>> model = load_model("blip_vqa")
            >>> samples = {
            ...     "image": torch.rand(2, 3, 480, 480),
            ...     "text_input": ["What is this?", "What is that?"],
            ...     "answer": ["cat", "cat", "dog"],
            ...     "weight": torch.tensor([1.0, 1.0, 1.0]),
            ...     "n_answers": torch.tensor([2, 1]),
            ... }
            >>> output = model(samples)
            >>> output.keys()
            odict_keys(['intermediate_output', 'loss'])
            >>> output.intermediate_output.keys()
            odict_keys(['image_embeds', 'encoder_output', 'decoder_output', 'decoder_labels'])
        ```
        """
        encoder_output, image_embeds = self.forward_encoder(samples)
        loss, decoder_output, decoder_targets = self.forward_decoder(
            samples=samples, encoder_out=encoder_output
        )

        return BlipOutput(
            loss=loss,
            intermediate_output=BlipIntermediateOutput(
                image_embeds=image_embeds,
                encoder_output=encoder_output,
                decoder_output=decoder_output,
                decoder_labels=decoder_targets,
            ),
        )

    def forward_encoder(self, samples):
        questions = samples["text_input"]
        questions = self.tokenizer(
            questions,
            padding="longest",
            truncation=True,
            max_length=self.max_txt_len,
            return_tensors="pt",
        ).to(self.device)
        questions.input_ids[:, 0] = self.tokenizer.enc_token_id
        samples.update({"tokenized_text": questions})

        image_embeds = self.visual_encoder.forward_features(samples["image"])
        encoder_output = self.text_encoder.forward_automask(
            tokenized_text=samples["tokenized_text"], visual_embeds=image_embeds
        )

        return encoder_output, image_embeds

    def forward_decoder(self, samples, encoder_out, **kwargs):
        answers = self.tokenizer(
            samples["answer"], padding="longest", return_tensors="pt"
        ).to(self.device)
        answers.input_ids[:, 0] = self.tokenizer.bos_token_id
        answer_targets = answers.input_ids.masked_fill(
            answers.input_ids == self.tokenizer.pad_token_id, -100
        )

        question_states = []
        question_atts = []

        question = samples["tokenized_text"]
        question_output = encoder_out

        for b, n in enumerate(samples["n_answers"]):
            question_states += [question_output.last_hidden_state[b]] * n
            question_atts += [question.attention_mask[b]] * n

        question_states = torch.stack(question_states, dim=0)
        question_atts = torch.stack(question_atts, dim=0)

        answer_output = self.text_decoder(
            answers.input_ids,
            attention_mask=answers.attention_mask,
            encoder_hidden_states=question_states,
            encoder_attention_mask=question_atts,
            labels=answer_targets,
            return_dict=True,
            reduction="none",
        )

        loss = samples["weight"] * answer_output.loss
        bsz = samples["image"].size(0)

        loss = loss.sum() / bsz

        return loss, answer_output, answer_targets

    def predict_answers(
        self,
        samples,
        num_beams=3,
        inference_method="rank",
        max_len=10,
        min_len=1,
        num_ans_candidates=128,
        answer_list=None,
        **kwargs
    ):
        """
        Args:
            samples (dict): A dictionary containing the following keys:
                - image (torch.Tensor): A tensor of shape (batch_size, 3, H, W). Default H=480, W=480.
                - text_input (str or [str]): String or a list of strings, each string is a question.
                                             The number of questions must be equal to the batch size. If a single string, will be converted to a list of string, with length 1 first.
            num_beams (int): Number of beams for beam search. 1 means no beam search.
            inference_method (str): Inference method. One of "rank", "generate".
                - If "rank", the model will return answers with the highest probability from the answer list.
                - If "generate", the model will generate answers.
            max_len (int): Maximum length of generated answers.
            min_len (int): Minimum length of generated answers.
            num_ans_candidates (int): Number of answer candidates, used to filter out answers with low probability.
            answer_list (list): A list of strings, each string is an answer.

        Returns:
            List: A list of strings, each string is an answer.

        Examples:
        ```python
            >>> from PIL import Image
            >>> from lavis.models import load_model_and_preprocess
            >>> model, vis_processors, txt_processors = load_model_and_preprocess("blip_vqa", "vqav2")
            >>> raw_image = Image.open("docs/data/merlion.png").convert("RGB")
            >>> question = "Which city is this photo taken?"
            >>> image = vis_processors["eval"](raw_image).unsqueeze(0)
            >>> question = txt_processors["eval"](question)
            >>> samples = {"image": image, "text_input": [question]}
            >>> answers = model.predict_answers(samples)
            >>> answers
            ['singapore']
            >>> answer_list = ["Singapore", "London", "Palo Alto", "Tokyo"]
            >>> answers = model.predict_answers(samples, answer_list=answer_list)
            >>> answers
            ['Singapore']
        ```
        """
        assert inference_method in [
            "rank",
            "generate",
        ], "Inference method must be one of 'rank' or 'generate', got {}.".format(
            inference_method
        )

        if isinstance(samples["text_input"], str):
            samples["text_input"] = [samples["text_input"]]

        assert len(samples["text_input"]) == samples["image"].size(
            0
        ), "The number of questions must be equal to the batch size."

        if inference_method == "generate":
            return self._generate_answers(
                samples, num_beams=num_beams, max_length=max_len, min_length=min_len
            )
        elif inference_method == "rank":
            assert answer_list is not None, "answer_list must be provided for ranking"

            num_ans_candidates = min(num_ans_candidates, len(answer_list))

            return self._rank_answers(
                samples, answer_list=answer_list, num_ans_candidates=num_ans_candidates
            )

    def _generate_answers(self, samples, num_beams=3, max_length=10, min_length=1):
        encoder_out, _ = self.forward_encoder(samples)

        question_output = encoder_out

        question_states = question_output.last_hidden_state.repeat_interleave(
            num_beams, dim=0
        )
        question_atts = torch.ones(question_states.size()[:-1], dtype=torch.long).to(
            self.device
        )

        model_kwargs = {
            "encoder_hidden_states": question_states,
            "encoder_attention_mask": question_atts,
        }

        bsz = samples["image"].size(0)
        bos_ids = torch.full(
            (bsz, 1), fill_value=self.tokenizer.bos_token_id, device=self.device
        )

        outputs = self.text_decoder.generate(
            input_ids=bos_ids,
            max_length=max_length,
            min_length=min_length,
            num_beams=num_beams,
            eos_token_id=self.tokenizer.sep_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
            **model_kwargs
        )

        # collect answers
        answers = []
        for output in outputs:
            answer = self.tokenizer.decode(output, skip_special_tokens=True)
            answers.append(answer)

        return answers

    def _rank_answers(self, samples, answer_list, num_ans_candidates):
        """
        Generate the first token of answers using decoder and select ${num_ans_candidates}
        most probable ones. Then select answers from answer list, which start with the probable tokens.
        Lastly, use the selected answers as the ground-truth labels for decoding and calculating LM loss.
        Return the answers that minimize the losses as result.

        """
        answer_candidates = self.tokenizer(
            answer_list, padding="longest", return_tensors="pt"
        ).to(self.device)
        answer_candidates.input_ids[:, 0] = self.tokenizer.bos_token_id

        answer_ids = answer_candidates.input_ids
        answer_atts = answer_candidates.attention_mask

        question_output, _ = self.forward_encoder(samples)
        question_states = question_output.last_hidden_state

        tokenized_question = samples["tokenized_text"]
        question_atts = tokenized_question.attention_mask

        num_ques = question_states.size(0)
        start_ids = answer_ids[0, 0].repeat(num_ques, 1)  # bos token

        start_output = self.text_decoder(
            start_ids,
            encoder_hidden_states=question_states,
            encoder_attention_mask=question_atts,
            return_dict=True,
            reduction="none",
        )
        logits = start_output.logits[:, 0, :]  # first token's logit

        # topk_probs: top-k probability
        # topk_ids: [num_question, k]
        answer_first_token = answer_ids[:, 1]
        prob_first_token = F.softmax(logits, dim=1).index_select(
            dim=1, index=answer_first_token
        )
        topk_probs, topk_ids = prob_first_token.topk(num_ans_candidates, dim=1)

        # answer input: [num_question*k, answer_len]
        input_ids = []
        input_atts = []
        for b, topk_id in enumerate(topk_ids):
            input_ids.append(answer_ids.index_select(dim=0, index=topk_id))
            input_atts.append(answer_atts.index_select(dim=0, index=topk_id))
        input_ids = torch.cat(input_ids, dim=0)
        input_atts = torch.cat(input_atts, dim=0)

        targets_ids = input_ids.masked_fill(
            input_ids == self.tokenizer.pad_token_id, -100
        )

        # repeat encoder's output for top-k answers
        question_states = tile(question_states, 0, num_ans_candidates)
        question_atts = tile(question_atts, 0, num_ans_candidates)

        output = self.text_decoder(
            input_ids,
            attention_mask=input_atts,
            encoder_hidden_states=question_states,
            encoder_attention_mask=question_atts,
            labels=targets_ids,
            return_dict=True,
            reduction="none",
        )

        log_probs_sum = -output.loss
        log_probs_sum = log_probs_sum.view(num_ques, num_ans_candidates)

        max_topk_ids = log_probs_sum.argmax(dim=1)
        max_ids = topk_ids[max_topk_ids >= 0, max_topk_ids]

        answers = [answer_list[max_id] for max_id in max_ids]

        return answers

    @classmethod
    def from_config(cls, cfg=None):
        image_encoder = VisionTransformerEncoder.from_config(cfg)

        # text encoder + multimodal encoder
        text_encoder = XBertEncoder.from_config(cfg)
        text_decoder = XBertLMHeadDecoder.from_config(cfg)

        max_txt_len = cfg.get("max_txt_len", 35)

        model = cls(
            image_encoder=image_encoder,
            text_encoder=text_encoder,
            text_decoder=text_decoder,
            max_txt_len=max_txt_len,
        )

        model.load_checkpoint_from_config(cfg)

        return model



import copy
import logging
import os
import os.path as osp
import io

try:
    import deepspeed
except Exception as e:
    print(e)
    print("deepspeed is not installed!!!")

from os.path import join


# try:
#     from petrel_client.client import Client
# except:
#     Client = None

import torch
from torch.utils.data import ConcatDataset, DataLoader

from ..dataset.resample_concat_dataset import ResampleConcatDataset
from ..models.backbones.internvideo2.pos_embed import interpolate_pos_embed_internvideo2_new
# from ..models.backbones.bert.tokenization_bert import BertTokenizer
from transformers import BertTokenizer
from ..utils.optimizer import create_optimizer
from ..utils.scheduler import create_scheduler
from ..utils.distributed import get_rank

logger = logging.getLogger(__name__)


def get_media_types(datasources):
    """get the media types for for all the dataloaders.

    Args:
        datasources (List): List of dataloaders or datasets.

    Returns: List. The media_types.

    """
    if isinstance(datasources[0], DataLoader):
        datasets = [dataloader.dataset for dataloader in datasources]
    else:
        datasets = datasources
    media_types = [
        dataset.datasets[0].media_type
        if isinstance(dataset, ConcatDataset) or isinstance(dataset, ResampleConcatDataset)
        else dataset.media_type
        for dataset in datasets
    ]

    return media_types


def setup_model(
    config, model_cls, add_decoder=False, pretrain=True, device="cuda"
):
    # logger.info("Creating model")
    config = copy.deepcopy(config)

    if "bert" in config.model.text_encoder.name:
        # logger.info(f"Using BertTokenizer: {config.model.text_encoder.pretrained}!")
        tokenizer = BertTokenizer.from_pretrained("google-bert/bert-large-uncased")#config.model.text_encoder.pretrained, local_files_only=True)
        model = model_cls(config=config, tokenizer=tokenizer, is_pretrain=pretrain)
    else:
        model = model_cls(config=config, is_pretrain=pretrain)
        tokenizer = model.tokenizer
        logger.info(f"Using model.tokenizer: {tokenizer}!")

    # if config.get('compile_model', False):
    #     torch.set_float32_matmul_precision('high')
    #     model = torch.compile(model)

    model = model.to(device)
    model_without_ddp = model


    # if hasattr(config, "deepspeed") and config.deepspeed.enable:
    #     # We move this to the back
    #     optimizer_params = create_optimizer(config.optimizer, model, return_group=True)
    #     scheduler = None
    #     scaler = None
    # else:
    #     if config.distributed:
    #         model = torch.nn.parallel.DistributedDataParallel(
    #             model,
    #             device_ids=[config.gpu],
    #             find_unused_parameters=find_unused_parameters,  # `False` for image-only task
    #         )

    #     optimizer = create_optimizer(config.optimizer, model)
    #     scaler = torch.cuda.amp.GradScaler(enabled=config.use_half_precision)  # This is never used actually if we fixed bf16
    #     scheduler = create_scheduler(config.scheduler, optimizer)


    # start_epoch = 0
    # global_step = 0

    # auto resume the latest checkpoint
    # if config.get("auto_resume", False):
        # logger.info("Auto resuming")
        # model_latest = join(config.output_dir, "ckpt_latest.pth")
        # model_best = join(config.output_dir, "ckpt_best.pth")
        # large_num = -1
        # for p in os.listdir(config.output_dir):
        #     if 'ckpt' in p:
        #         num = p.split('_')[1].split('.')[0]
        #         if str.isnumeric(num):
        #             if int(num) > large_num:
        #                 large_num = int(num)
        # if large_num != -1:
        #     model_latest = join(config.output_dir, f"ckpt_{large_num:02d}.pth")
        # if osp.isfile(model_latest):
        #     config.pretrained_path = model_latest
        #     config.resume = True
        # elif osp.isfile(model_best):
        #     config.pretrained_path = model_best
        #     config.resume = True
        # else:
        #     logger.info(f"Not found checkpoint in {config.output_dir}")


    if (config.pretrained_path.strip() and (osp.isfile(config.pretrained_path)) or "s3://" in config.pretrained_path):
        # if Client is not None:
        #     client = Client()
        #     with io.BytesIO(client.get(config.pretrained_path)) as buffer:
        #         checkpoint = torch.load(buffer, map_location="cpu")
        # else:
        checkpoint = torch.load(config.pretrained_path, map_location="cpu")
        # logger.info(f"Loading checkpoint from {config.pretrained_path}")
        try:
            if "model" in checkpoint.keys():
                state_dict = checkpoint["model"]
            else:
                state_dict = checkpoint["module"] # This is a deepspeed stage 1 model
        except:  
            state_dict = checkpoint

        if config.get('origin_num_frames', None) is not None:
            # logger.info(f"interpolate_pos_embed_internvideo2 (origin_num_frames={config.origin_num_frames})!!!")
            a = len(state_dict)
            interpolate_pos_embed_internvideo2_new(state_dict, model_without_ddp.vision_encoder, orig_t_size=config.origin_num_frames)
            assert a == len(state_dict), state_dict.keys()

        # if config.resume:
        #     assert not (hasattr(config, "deepspeed") and config.deepspeed.enable), "Deepspeed should run here!!!"
        #     optimizer.load_state_dict(checkpoint["optimizer"])
        #     scheduler.load_state_dict(checkpoint["scheduler"])
        #     scaler.load_state_dict(checkpoint["scaler"])
        #     if 'local_step' in checkpoint.keys():
        #         start_epoch = checkpoint['epoch']
        #     else:
        #         start_epoch = checkpoint["epoch"] + 1
        #     global_step = checkpoint["global_step"]

        elif not pretrain:  # downstream init from pretrained ckpt 


            if not config.evaluate or config.get("zero_shot", False):  # finetuning from a pretrained weights.
                if add_decoder:
                    logger.info("Init new decoder with encoder!!!")
                for key in list(state_dict.keys()):
                    if "text_encoder.bert" in key:
                        encoder_key = key.replace("bert.", "") 
                        state_dict[encoder_key] = state_dict[key]
                        if not add_decoder:
                            del state_dict[key]

                    # init text decoder as multimodal encoder (last 6 layers of model.text_encoder)
                    # only for generation tasks like VQA
                    if add_decoder and "text_encoder.bert" in key:
                        if "layer" in key:
                            encoder_keys = key.split(".")
                            layer_num = int(encoder_keys[4])
                            if layer_num < config.model.text_encoder.fusion_layer:
                                del state_dict[key]
                                continue
                            else:
                                decoder_layer_num = layer_num - config.model.text_encoder.fusion_layer
                                encoder_keys[4] = str(decoder_layer_num)
                                encoder_key = ".".join(encoder_keys)
                        else:
                            encoder_key = key
                        decoder_key = encoder_key.replace("text_encoder", "text_decoder")
                        state_dict[decoder_key] = state_dict[key]
                        del state_dict[key]


        msg = model.load_state_dict(state_dict, strict=True)
        # logger.info(msg)
        # logger.info(f"Loaded checkpoint from {config.pretrained_path}")
    else:
        raise ValueError("No pretrained checkpoint provided!!!")
        # if not config.resume:
        #     assert not config.evaluate, "No available pretrained checkpoint provided!!!"
        #     assert config.pretrained_path == "", config.pretrained_path
        #     logger.warning("No available pretrained checkpoint provided, training from scratch.")


    # if hasattr(config, "deepspeed") and config.deepspeed.enable:
    #     logger.info(f'Use deepspeed to initialize model (resume={config.resume}) !!!')
    #     model = model_without_ddp

    #     model, optimizer, _, _ = deepspeed.initialize(
    #         args=config, model=model, model_parameters=optimizer_params, dist_init_required=not config.distributed,
    #         lr_scheduler=lambda opt: create_scheduler(config.scheduler, opt)
    #     )


    #     if config.resume:
    #         logger.info(f'Resume deepspeed ckpt from {config.output_dir}, tag={config.pretrained_path}, load_module_strict={config.get("load_module_strict", True)}, load_lr_scheduler_states={config.get("load_lr_scheduler_states", True)}!!!')
    #         _, client_states = model.load_checkpoint(config.output_dir, tag=config.pretrained_path,  load_module_strict=config.get("load_module_strict", True), load_lr_scheduler_states=config.get("load_lr_scheduler_states", True))
    #         logger.info(client_states)
    #         if 'local_step' in client_states.keys():
    #             start_epoch = client_states['epoch']
    #         else:
    #             start_epoch = client_states['epoch'] + 1
    #         global_step = client_states['global_step']


    # logger.info(f"Cuda memory after create model: {torch.cuda.memory_allocated() // 1024**2}M, Max mem: {torch.cuda.max_memory_allocated() // 1024**2}M start_epoch={start_epoch}, global_step={global_step}")
    # print(f"\033[31m Cuda memory after create model: {torch.cuda.memory_allocated() // 1024**2}M, Max mem: {torch.cuda.max_memory_allocated() // 1024**2}M start_epoch={start_epoch}, global_step={global_step}\033[0m")

    return (
        model,
        # model_without_ddp,
        # optimizer,
        # scheduler,
        # scaler,
        tokenizer,
        # start_epoch,
        # global_step,
    )



"""Tests for deepvqa module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_deepvqa_basics():
    from ayase.modules.deepvqa import DeepVQAModule
    _test_module_basics(DeepVQAModule, "deepvqa")

def test_deepvqa_no_reference(image_sample):
    from ayase.modules.deepvqa import DeepVQAModule
    m = DeepVQAModule()
    result = m.process(image_sample)
    assert result is image_sample



"""Tests for inception_score module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_inception_score_basics():
    from ayase.modules.inception_score import InceptionScoreModule
    _test_module_basics(InceptionScoreModule, "inception_score")

def test_inception_score_image(image_sample):
    from ayase.modules.inception_score import InceptionScoreModule
    image_sample.quality_metrics = QualityMetrics()
    m = InceptionScoreModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_inception_score_video(video_sample):
    from ayase.modules.inception_score import InceptionScoreModule
    video_sample.quality_metrics = QualityMetrics()
    m = InceptionScoreModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



import time
from joblib import Parallel,delayed
import numpy as np
import cv2
import queue
import glob
import os
import time
import scipy.ndimage
import joblib
import sys
import matplotlib.pyplot as plt
import niqe 
import save_stats
from numba import jit,prange
import argparse
try:
    from pypapi import events, papi_high as high
except ImportError:
    events = None
    high = None

parser = argparse.ArgumentParser(description='Generate ChipQA features from a folder of videos and store them')
parser.add_argument('input_folder',help='Folder containing input videos')
parser.add_argument('results_folder',help='Folder where features are stored')

args = parser.parse_args()
C=1
def gen_gauss_window(lw, sigma):
    sd = np.float32(sigma)
    lw = int(lw)
    weights = [0.0] * (2 * lw + 1)
    weights[lw] = 1.0
    sum = 1.0
    sd *= sd
    for ii in range(1, lw + 1):
        tmp = np.exp(-0.5 * np.float32(ii * ii) / sd)
        weights[lw + ii] = tmp
        weights[lw - ii] = tmp
        sum += 2.0 * tmp
    for ii in range(2 * lw + 1):
        weights[ii] /= sum
    return weights
def compute_image_mscn_transform(image, C=1, avg_window=None, extend_mode='constant'):
    if avg_window is None:
      avg_window = gen_gauss_window(3, 7.0/6.0)
    assert len(np.shape(image)) == 2
    h, w = np.shape(image)
    mu_image = np.zeros((h, w), dtype=np.float32)
    var_image = np.zeros((h, w), dtype=np.float32)
    image = np.array(image).astype('float32')
    scipy.ndimage.correlate1d(image, avg_window, 0, mu_image, mode=extend_mode)
    scipy.ndimage.correlate1d(mu_image, avg_window, 1, mu_image, mode=extend_mode)
    scipy.ndimage.correlate1d(image**2, avg_window, 0, var_image, mode=extend_mode)
    scipy.ndimage.correlate1d(var_image, avg_window, 1, var_image, mode=extend_mode)
    var_image = np.sqrt(np.abs(var_image - mu_image**2))
    return (image - mu_image)/(var_image + C), var_image, mu_image

def spatiotemporal_mscn(img_buffer,avg_window,extend_mode='mirror'):
    st_mean = np.zeros((img_buffer.shape))
    scipy.ndimage.correlate1d(img_buffer, avg_window, 0, st_mean, mode=extend_mode)
    return st_mean

@jit(nopython=True)
def find_sts_locs(sts_slope,cy,cx,step,h,w):
    if(np.abs(sts_slope)<1):
        x_sts = np.arange(cx-int((step-1)/2),cx+int((step-1)/2)+1)
        y = (cy-(x_sts-cx)*sts_slope).astype(np.int64)
        y_sts = np.asarray([y[j] if y[j]<h else h-1 for j in range(step)])
    else:
        #        print(np.abs(sts_slope))
        y_sts = np.arange(cy-int((step-1)/2),cy+int((step-1)/2)+1)
        x= ((-y_sts+cy)/sts_slope+cx).astype(np.int64)
        x_sts = np.asarray([x[j] if x[j]<w else w-1 for j in range(step)]) 
    return x_sts,y_sts


@jit(nopython=True)
def find_kurtosis_slice(Y3d_mscn,cy,cx,rst,rct,theta,h,step):
    st_kurtosis = np.zeros((len(theta),))
    data = np.zeros((len(theta),step**2))
    for index,t in enumerate(theta):
        rsin_theta = rst[:,index]
        rcos_theta  =rct[:,index]
        x_sts,y_sts = cx+rcos_theta,cy+rsin_theta
        
        data[index,:] =Y3d_mscn[:,y_sts*h+x_sts].flatten() 
        data_mu4 = np.mean((data[index,:]-np.mean(data[index,:]))**4)
        data_var = np.var(data[index,:])
        st_kurtosis[index] = data_mu4/(data_var**2+1e-4)
    idx = (np.abs(st_kurtosis - 3)).argmin()
    
    data_slice = data[idx,:]
    return data_slice,st_kurtosis[idx]-3


def find_kurtosis_sts(img_buffer,grad_img_buffer,step,cy,cx,rst,rct,theta):

    h, w = img_buffer[step-1].shape[:2]
    Y3d_mscn = np.reshape(img_buffer.copy(),(step,-1))
    gradY3d_mscn = np.reshape(grad_img_buffer.copy(),(step,-1))
    sts= [find_kurtosis_slice(Y3d_mscn,cy[i],cx[i],rst,rct,theta,h,step) for i in range(len(cy))]
    sts_grad= [find_kurtosis_slice(gradY3d_mscn,cy[i],cx[i],rst,rct,theta,h,step) for i in range(len(cy))]

    st_data = [sts[i][0] for i in range(len(sts))]
    st_deviation = [sts[i][1] for i in range(len(sts))]
    st_grad_data = [sts_grad[i][0] for i in range(len(sts_grad))]
    st_grad_dev = [sts_grad[i][1] for i in range(len(sts_grad))]
    return st_data,np.asarray(st_deviation),st_grad_data,np.asarray(st_grad_dev)



def sts_fromfilename(i,filenames,results_folder):
    filename = filenames[i]
    name = os.path.basename(filename)
    print(name) 
    filename_out =os.path.join(results_folder,os.path.splitext(name)[0]+'.z')
    st_time_length = 5
    t = np.arange(0,st_time_length)
    a=0.25
    avg_window = t*(1-a*t)*np.exp(-2*a*t)
    avg_window = np.flip(avg_window)
    cap = cv2.VideoCapture(filename)
    count=1
    ret, prev = cap.read()
    print(ret)
                    #percent by which the image is resized
    scale_percent = 0.5
#
    theta = np.arange(0,np.pi,np.pi/6)
    print(len(theta))
    ct = np.cos(theta)
    st = np.sin(theta)
    lower_r = int((st_time_length+1)/2)-1
    higher_r = int((st_time_length+1)/2)
    r = np.arange(-lower_r,higher_r)
    rct = np.round(np.outer(r,ct))
    rst = np.round(np.outer(r,st))
    rct = rct.astype(np.int32)
    rst = rst.astype(np.int32)


    prevY = cv2.cvtColor(prev, cv2.COLOR_BGR2GRAY)
    prevY = prevY.astype(np.float32)
    h,w = prev.shape[0],prev.shape[1]
    print(h,w)
    if(h>w):
        h_temp = h
        h=w
        w = h_temp

    # dsize
    dsize = (int(scale_percent*h),int(scale_percent*w))
    print(h,w,dsize)

    step = st_time_length
    cy, cx = np.mgrid[step:h-step*4:step*4, step:w-step*4:step*4].reshape(2,-1).astype(int) # these will be the centers of each block
    dcy, dcx = np.mgrid[step:dsize[0]-step*4:step*4, step:dsize[1]-step*4:step*4].reshape(2,-1).astype(int) # these will be the centers of each block

    
    prevY_down = cv2.resize(prevY,(dsize[1],dsize[0]),interpolation=cv2.INTER_CUBIC)

    img_buffer = np.zeros((st_time_length,prevY.shape[0],prevY.shape[1]))
    grad_img_buffer = np.zeros((st_time_length,prevY.shape[0],prevY.shape[1]))
    down_img_buffer =np.zeros((st_time_length,prevY_down.shape[0],prevY_down.shape[1]))
    graddown_img_buffer =np.zeros((st_time_length,prevY_down.shape[0],prevY_down.shape[1]))


    gradient_x = cv2.Sobel(prevY,ddepth=-1,dx=1,dy=0)
    gradient_y = cv2.Sobel(prevY,ddepth=-1,dx=0,dy=1)
    gradient_mag = np.sqrt(gradient_x**2+gradient_y**2)    

    
    gradient_x_down = cv2.Sobel(prevY_down,ddepth=-1,dx=1,dy=0)
    gradient_y_down = cv2.Sobel(prevY_down,ddepth=-1,dx=0,dy=1)
    gradient_mag_down = np.sqrt(gradient_x_down**2+gradient_y_down**2)    
    i = 0

    Y_mscn,_,_ = compute_image_mscn_transform(prevY)
    dY_mscn,_,_ = compute_image_mscn_transform(prevY_down)
    gradY_mscn,_,_ = compute_image_mscn_transform(gradient_mag)
    dgradY_mscn,_,_ = compute_image_mscn_transform(gradient_mag_down)

    img_buffer[i,:,:] = Y_mscn
    down_img_buffer[i,:,:]= dY_mscn
    grad_img_buffer[i,:,:] =gradY_mscn 
    graddown_img_buffer[i,:,:]=dgradY_mscn 
    i = i+1

    r1 = len(np.arange(step,h-step*4,step*4)) 
    r2 = len(np.arange(step,w-step*4,step*4)) 
    dr1 = len(np.arange(step,dsize[0]-step*4,step*4)) 
    dr2 = len(np.arange(step,dsize[1]-step*4,step*4)) 
    
    head, tail = os.path.split(filename)

    

    
    spat_list = []
    X_list = []
    spatavg_list = []
    feat_sd_list =  []
    sd_list= []
    
    j=0
    total_time = 0
    while(True):
        # try:
            # 
        j = j+1
        
        # uncomment for FLOPS
        #high.start_counters([events.PAPI_FP_OPS,])

        
        ret,bgr = cap.read()
        count=count+1
        print(count)
        if(ret==False):
            count=count-1
            break
        lab = cv2.cvtColor(bgr, cv2.COLOR_BGR2LAB)
        lab = lab.astype(np.float32)
        chroma_feats = save_stats.chroma_feats(lab)

        Y = cv2.cvtColor(bgr, cv2.COLOR_BGR2GRAY)
        Y = Y.astype(np.float32)
        Y_down = cv2.resize(Y,(dsize[1],dsize[0]),interpolation=cv2.INTER_CUBIC)
#
        gradient_x = cv2.Sobel(Y,ddepth=-1,dx=1,dy=0)
        gradient_y = cv2.Sobel(Y,ddepth=-1,dx=0,dy=1)
        gradient_mag = np.sqrt(gradient_x**2+gradient_y**2)    

        
        gradient_x_down = cv2.Sobel(Y_down,ddepth=-1,dx=1,dy=0)
        gradient_y_down = cv2.Sobel(Y_down,ddepth=-1,dx=0,dy=1)
        gradient_mag_down = np.sqrt(gradient_x_down**2+gradient_y_down**2)    


        Y_mscn,Ysigma,_ = compute_image_mscn_transform(Y)
        dY_mscn,dYsigma,_ = compute_image_mscn_transform(Y_down)

        gradY_mscn,_,_ = compute_image_mscn_transform(gradient_mag)
        dgradY_mscn,_,_ = compute_image_mscn_transform(gradient_mag_down)

        gradient_feats = save_stats.extract_secondord_feats(gradY_mscn)
        gdown_feats = save_stats.extract_secondord_feats(dgradY_mscn)
        gfeats = np.concatenate((gradient_feats,gdown_feats),axis=0)

        
        Ysigma_mscn,_,_= compute_image_mscn_transform(Ysigma)
        dYsigma_mscn,_,_= compute_image_mscn_transform(dYsigma)

        sigma_feats = save_stats.stat_feats(Ysigma_mscn)
        dsigma_feats = save_stats.stat_feats(dYsigma_mscn)


        feats = np.concatenate((chroma_feats,gfeats,sigma_feats,dsigma_feats),axis=0)

        feat_sd_list.append(feats)
        spatavg_list.append(feats)

        
        img_buffer[i,:,:] = Y_mscn
        down_img_buffer[i,:,:]= dY_mscn
        grad_img_buffer[i,:,:] =gradY_mscn 
        graddown_img_buffer[i,:,:]=dgradY_mscn 
        i=i+1
#

        if (i>=st_time_length): 

            Y3d_mscn = spatiotemporal_mscn(img_buffer,avg_window)
            Ydown_3d_mscn = spatiotemporal_mscn(down_img_buffer,avg_window)
            grad3d_mscn = spatiotemporal_mscn(grad_img_buffer,avg_window)
            graddown3d_mscn = spatiotemporal_mscn(graddown_img_buffer,avg_window)
            spat_feats = niqe.compute_niqe_features(Y)

            sd_feats = np.std(feat_sd_list,axis=0)
            sd_list.append(sd_feats)
            feat_sd_list = []

            sts,st_deviation,sts_grad,sts_grad_deviation = find_kurtosis_sts(Y3d_mscn,grad3d_mscn,step,cy,cx,rst,rct,theta)
            dsts,dsts_deviation,dsts_grad,dsts_grad_deviation = find_kurtosis_sts(Ydown_3d_mscn,graddown3d_mscn,step,dcy,dcx,rst,rct,theta)
            sts_arr = np.reshape(sts,(r1*st_time_length,r2*st_time_length)) 
            sts_grad = np.reshape(sts_grad,(r1*st_time_length,r2*st_time_length))
            print()
            dsts_arr = np.reshape(dsts,(dr1*st_time_length,dr2*st_time_length)) #(int((int(dsize[0]/20)*20-step*4)/4),int((int(dsize[1]/20)*20-step*4)/4)))
            dsts_grad = np.reshape(dsts_grad,(dr1*st_time_length,dr2*st_time_length))#(int((int(dsize[0]/20)*20-step*4)/4),int((int(dsize[1]/20)*20-step*4)/4)))
            feats =  save_stats.brisque(sts_arr)
            grad_feats = save_stats.brisque(sts_grad)
            
            dfeats =  save_stats.brisque(dsts_arr)
            dgrad_feats = save_stats.brisque(dsts_grad)


            allst_feats = np.concatenate((spat_feats,feats,dfeats,grad_feats,dgrad_feats),axis=0)
            print(allst_feats.shape)
            X_list.append(allst_feats)


            img_buffer = np.zeros((st_time_length,prevY.shape[0],prevY.shape[1]))
            grad_img_buffer = np.zeros((st_time_length,prevY.shape[0],prevY.shape[1]))
            down_img_buffer =np.zeros((st_time_length,prevY_down.shape[0],prevY_down.shape[1]))
            graddown_img_buffer =np.zeros((st_time_length,prevY_down.shape[0],prevY_down.shape[1]))
            i=0
#            x=high.stop_counters()
#        print(x,"is the number of flops")

    X1 = np.average(spatavg_list,axis=0)
    X2 = np.average(sd_list,axis=0)
    X3 = np.average(X_list,axis=0)
    X = np.concatenate((X1,X2,X3),axis=0)
    train_dict = {"features":X}
    filename_out =os.path.join(os.path.splitext(name)[0]+'.z')
    joblib.dump(train_dict,os.path.join(results_folder,filename_out))
    return


def sts_fromvid(args):
    filenames = glob.glob(os.path.join(args.input_folder,'*.mp4'))
    print(sorted(filenames))
    filenames = sorted(filenames)
    flag = 0
    Parallel(n_jobs=1)(delayed(sts_fromfilename)(i,filenames,args.results_folder) for i in range(len(filenames)))
#    sts_fromfilename(34,filenames,args.results_folder)
             



    return


def main():
    args = parser.parse_args()
    sts_fromvid(args)


if __name__ == '__main__':
    # print(__doc__)
    main()
    




from pathlib import Path
from unittest.mock import MagicMock

import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample, VideoMetadata


def test_cambi_basics():
    from ayase.modules.cambi import CAMBIModule
    from .conftest import _test_module_basics

    _test_module_basics(CAMBIModule, "cambi")


def test_cambi_image(image_sample):
    from ayase.modules.cambi import CAMBIModule

    m = CAMBIModule()
    m._ml_available = True
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.cambi is None


def test_cambi_no_ffmpeg(video_sample):
    from ayase.modules.cambi import CAMBIModule

    m = CAMBIModule()
    assert m._ml_available is False
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.cambi is None


def test_xpsnr_basics():
    from ayase.modules.xpsnr import XPSNRModule
    from .conftest import _test_module_basics

    _test_module_basics(XPSNRModule, "xpsnr")


def test_xpsnr_no_reference(video_sample):
    from ayase.modules.xpsnr import XPSNRModule

    m = XPSNRModule()
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.xpsnr is None


def test_xpsnr_image(image_sample):
    from ayase.modules.xpsnr import XPSNRModule

    m = XPSNRModule()
    m._ml_available = True
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.xpsnr is None


def test_vmaf_phone_basics():
    from ayase.modules.vmaf_phone import VMAFPhoneModule
    from .conftest import _test_module_basics

    _test_module_basics(VMAFPhoneModule, "vmaf_phone")


def test_vmaf_phone_no_reference(video_sample):
    from ayase.modules.vmaf_phone import VMAFPhoneModule

    m = VMAFPhoneModule()
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.vmaf_phone is None


def test_vmaf_4k_basics():
    from ayase.modules.vmaf_4k import VMAF4KModule
    from .conftest import _test_module_basics

    _test_module_basics(VMAF4KModule, "vmaf_4k")


def test_vmaf_4k_no_reference(video_sample):
    from ayase.modules.vmaf_4k import VMAF4KModule

    m = VMAF4KModule()
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.vmaf_4k is None


def test_visqol_basics():
    from ayase.modules.visqol import ViSQOLModule
    from .conftest import _test_module_basics

    _test_module_basics(ViSQOLModule, "visqol")


def test_visqol_image(image_sample):
    from ayase.modules.visqol import ViSQOLModule

    m = ViSQOLModule()
    m._ml_available = True
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.visqol is None


def test_visqol_no_backend(video_sample):
    from ayase.modules.visqol import ViSQOLModule

    m = ViSQOLModule()
    assert m._ml_available is False
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.visqol is None


def test_dnsmos_basics():
    from ayase.modules.dnsmos import DNSMOSModule
    from .conftest import _test_module_basics

    _test_module_basics(DNSMOSModule, "dnsmos")


def test_dnsmos_image(image_sample):
    from ayase.modules.dnsmos import DNSMOSModule

    m = DNSMOSModule()
    m._ml_available = True
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.dnsmos_overall is None


def test_dnsmos_no_backend(video_sample):
    from ayase.modules.dnsmos import DNSMOSModule

    m = DNSMOSModule()
    assert m._ml_available is False
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.dnsmos_overall is None


def test_pu_metrics_basics():
    from ayase.modules.pu_metrics import PUMetricsModule
    from .conftest import _test_module_basics

    _test_module_basics(PUMetricsModule, "pu_metrics")


def test_pu_metrics_no_reference(image_sample):
    from ayase.modules.pu_metrics import PUMetricsModule

    m = PUMetricsModule()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.pu_psnr is None


def test_pu_metrics_setup():
    from ayase.modules.pu_metrics import PUMetricsModule

    m = PUMetricsModule()
    m.setup()
    assert m._ml_available is True


def test_pu21_encode():
    from ayase.modules.pu_metrics import _pu21_encode

    linear = np.array([0.0, 0.5, 1.0])
    encoded = _pu21_encode(linear)
    assert len(encoded) == 3
    assert encoded[0] < encoded[1] < encoded[2]


def test_pu_metrics_compute():
    from ayase.modules.pu_metrics import PUMetricsModule

    m = PUMetricsModule()
    m.setup()
    ref = np.full((64, 64, 3), 5000.0, dtype=np.float32)
    dist = ref.copy()
    dist[:32, :32] = 7000.0
    psnr = m._compute_pu_psnr(ref, dist)
    assert psnr > 0
    ssim = m._compute_pu_ssim(ref, dist)
    assert 0.0 <= ssim <= 1.0


def test_pu_metrics_identical():
    from ayase.modules.pu_metrics import PUMetricsModule

    m = PUMetricsModule()
    ref = np.full((64, 64, 3), 5000.0, dtype=np.float32)
    psnr = m._compute_pu_psnr(ref, ref)
    assert psnr >= 90.0


def test_hdr_metadata_basics():
    from ayase.modules.hdr_metadata import HDRMetadataModule
    from .conftest import _test_module_basics

    _test_module_basics(HDRMetadataModule, "hdr_metadata")


def test_hdr_metadata_image(image_sample):
    from ayase.modules.hdr_metadata import HDRMetadataModule

    m = HDRMetadataModule()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.max_fall is None


def test_hdr_metadata_video(video_sample):
    from ayase.modules.hdr_metadata import HDRMetadataModule

    m = HDRMetadataModule()
    result = m.process(video_sample)
    assert result.quality_metrics is not None
    assert result.quality_metrics.max_fall is not None
    assert result.quality_metrics.max_fall >= 0
    assert result.quality_metrics.max_cll is not None
    assert result.quality_metrics.max_cll >= 0


def test_bt709_luminance():
    from ayase.modules.hdr_metadata import _bt709_luminance

    white = np.array([[[255, 255, 255]]], dtype=np.float32)
    lum = _bt709_luminance(white)
    assert abs(lum[0, 0] - 255.0) < 1.0


def test_hdr_vdp_basics():
    from ayase.modules.hdr_vdp import HDRVDPModule
    from .conftest import _test_module_basics

    _test_module_basics(HDRVDPModule, "hdr_vdp")


def test_hdr_vdp_no_reference(image_sample):
    from ayase.modules.hdr_vdp import HDRVDPModule

    m = HDRVDPModule()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.hdr_vdp is None


def test_hdr_vdp_setup():
    from ayase.modules.hdr_vdp import HDRVDPModule

    m = HDRVDPModule()
    m.setup()
    assert m._ml_available is True
    assert m._backend == "approx"


def test_hdr_vdp_reference(tmp_dir):
    from ayase.modules.hdr_vdp import HDRVDPModule

    m = HDRVDPModule()
    m.setup()
    ref_img = np.full((64, 64, 3), 128, dtype=np.uint8)
    dist_img = ref_img.copy()
    dist_img[:32, :32] = 200
    ref_path = tmp_dir / "ref.png"
    dist_path = tmp_dir / "dist.png"
    import cv2

    cv2.imwrite(str(ref_path), ref_img)
    cv2.imwrite(str(dist_path), dist_img)
    score = m.compute_reference_score(dist_path, ref_path)
    assert score is not None
    assert score > 0


def test_delta_ictcp_basics():
    from ayase.modules.delta_ictcp import DeltaICtCpModule
    from .conftest import _test_module_basics

    _test_module_basics(DeltaICtCpModule, "delta_ictcp")


def test_delta_ictcp_no_reference(image_sample):
    from ayase.modules.delta_ictcp import DeltaICtCpModule

    m = DeltaICtCpModule()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.delta_ictcp is None


def test_delta_ictcp_pq():
    from ayase.modules.delta_ictcp import _linear_to_pq

    linear = np.array([0.0, 0.25, 0.5, 0.75, 1.0])
    pq = _linear_to_pq(linear)
    for i in range(len(pq) - 1):
        assert pq[i] <= pq[i + 1]


def test_delta_ictcp_identical(tmp_dir):
    from ayase.modules.delta_ictcp import DeltaICtCpModule

    m = DeltaICtCpModule()
    m._ml_available = True
    img = np.full((64, 64, 3), 128, dtype=np.uint8)
    p = tmp_dir / "img.png"
    import cv2

    cv2.imwrite(str(p), img)
    score = m.compute_reference_score(p, p)
    assert score is not None
    assert score < 0.01


def test_delta_ictcp_different(tmp_dir):
    from ayase.modules.delta_ictcp import DeltaICtCpModule

    m = DeltaICtCpModule()
    m._ml_available = True
    ref = np.full((64, 64, 3), 128, dtype=np.uint8)
    dist = np.full((64, 64, 3), 50, dtype=np.uint8)
    ref_path = tmp_dir / "ref.png"
    dist_path = tmp_dir / "dist.png"
    import cv2

    cv2.imwrite(str(ref_path), ref)
    cv2.imwrite(str(dist_path), dist)
    score = m.compute_reference_score(dist_path, ref_path)
    assert score is not None
    assert score > 0


def test_strred_basics():
    from ayase.modules.strred import STRREDModule
    from .conftest import _test_module_basics

    _test_module_basics(STRREDModule, "strred")


def test_strred_no_reference(video_sample):
    from ayase.modules.strred import STRREDModule

    m = STRREDModule()
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.strred is None


def test_strred_setup():
    from ayase.modules.strred import STRREDModule

    m = STRREDModule()
    m.setup()
    assert m._ml_available is True
    assert m._backend in ("skvideo", "approx")


def test_cgvqm_basics():
    from ayase.modules.cgvqm import CGVQMModule
    from .conftest import _test_module_basics

    _test_module_basics(CGVQMModule, "cgvqm")


def test_cgvqm_no_reference(image_sample):
    from ayase.modules.cgvqm import CGVQMModule

    m = CGVQMModule()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.cgvqm is None


def test_cgvqm_setup():
    from ayase.modules.cgvqm import CGVQMModule

    m = CGVQMModule()
    m.setup()
    assert m._ml_available is True
    assert m._backend == "approx"


def test_cgvqm_reference(tmp_dir):
    from ayase.modules.cgvqm import CGVQMModule

    m = CGVQMModule()
    m.setup()
    ref_img = np.full((64, 64, 3), 128, dtype=np.uint8)
    dist_img = ref_img.copy()
    dist_img[:32, :32] = 200
    ref_path = tmp_dir / "ref.png"
    dist_path = tmp_dir / "dist.png"
    import cv2

    cv2.imwrite(str(ref_path), ref_img)
    cv2.imwrite(str(dist_path), dist_img)
    score = m.compute_reference_score(dist_path, ref_path)
    assert score is not None
    assert 0.0 <= score <= 100.0


def test_ciede2000_basics():
    from ayase.modules.ciede2000 import CIEDE2000Module
    from .conftest import _test_module_basics

    _test_module_basics(CIEDE2000Module, "ciede2000")


def test_ciede2000_no_reference(image_sample):
    from ayase.modules.ciede2000 import CIEDE2000Module

    m = CIEDE2000Module()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.ciede2000 is None


def test_ciede2000_identical(tmp_dir):
    from ayase.modules.ciede2000 import CIEDE2000Module

    m = CIEDE2000Module()
    img = np.full((64, 64, 3), 128, dtype=np.uint8)
    p = tmp_dir / "img.png"
    import cv2

    cv2.imwrite(str(p), img)
    score = m.compute_reference_score(p, p)
    assert score is not None
    assert score < 0.1


def test_ciede2000_different(tmp_dir):
    from ayase.modules.ciede2000 import CIEDE2000Module

    m = CIEDE2000Module()
    ref = np.full((64, 64, 3), 128, dtype=np.uint8)
    dist = np.full((64, 64, 3), 50, dtype=np.uint8)
    ref_path = tmp_dir / "ref.png"
    dist_path = tmp_dir / "dist.png"
    import cv2

    cv2.imwrite(str(ref_path), ref)
    cv2.imwrite(str(dist_path), dist)
    score = m.compute_reference_score(dist_path, ref_path)
    assert score is not None
    assert score > 0


def test_ciede2000_vectorized():
    from ayase.modules.ciede2000 import _ciede2000_pixel

    lab1 = np.array([[[50.0, 0.0, 0.0]]])
    lab2 = np.array([[[50.0, 0.0, 0.0]]])
    de = _ciede2000_pixel(lab1, lab2)
    assert de[0, 0] < 0.01


def test_psnr_hvs_basics():
    from ayase.modules.psnr_hvs import PSNRHVSModule
    from .conftest import _test_module_basics

    _test_module_basics(PSNRHVSModule, "psnr_hvs")


def test_psnr_hvs_no_reference(image_sample):
    from ayase.modules.psnr_hvs import PSNRHVSModule

    m = PSNRHVSModule()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.psnr_hvs is None


def test_psnr_hvs_setup():
    from ayase.modules.psnr_hvs import PSNRHVSModule

    m = PSNRHVSModule()
    m.setup()
    assert m._ml_available is True


def test_psnr_hvs_identical(tmp_dir):
    from ayase.modules.psnr_hvs import PSNRHVSModule

    m = PSNRHVSModule()
    m.setup()
    img = np.full((64, 64, 3), 128, dtype=np.uint8)
    p = tmp_dir / "img.png"
    import cv2

    cv2.imwrite(str(p), img)
    score = m.compute_reference_score(p, p)
    assert score is not None
    assert score >= 90.0


def test_psnr_hvs_different(tmp_dir):
    from ayase.modules.psnr_hvs import PSNRHVSModule

    m = PSNRHVSModule()
    m.setup()
    ref = np.full((64, 64, 3), 128, dtype=np.uint8)
    dist = ref.copy()
    dist[:32, :32] = 200
    ref_path = tmp_dir / "ref.png"
    dist_path = tmp_dir / "dist.png"
    import cv2

    cv2.imwrite(str(ref_path), ref)
    cv2.imwrite(str(dist_path), dist)
    score = m.compute_reference_score(dist_path, ref_path)
    assert score is not None
    assert score > 0
    assert score < 100


def test_bd_rate_basics():
    from ayase.modules.bd_rate import BDRateModule
    from .conftest import _test_module_basics

    _test_module_basics(BDRateModule, "bd_rate")


def test_bd_rate_no_metadata(video_sample):
    from ayase.modules.bd_rate import BDRateModule

    m = BDRateModule()
    assert len(m._feature_cache) == 0


def test_bd_rate_function():
    from ayase.modules.bd_rate import _bd_rate

    rates = [500, 1000, 2000, 4000]
    quality = [30.0, 35.0, 40.0, 45.0]
    bd = _bd_rate(rates, quality, rates, quality)
    assert abs(bd) < 1.0


def test_bd_rate_better_codec():
    from ayase.modules.bd_rate import _bd_rate

    rates1 = [500, 1000, 2000, 4000]
    quality1 = [30.0, 35.0, 40.0, 45.0]
    rates2 = [350, 700, 1400, 2800]
    quality2 = [30.0, 35.0, 40.0, 45.0]
    bd = _bd_rate(rates1, quality1, rates2, quality2)
    assert bd < 0


def test_p1203_basics():
    from ayase.modules.p1203 import P1203Module
    from .conftest import _test_module_basics

    _test_module_basics(P1203Module, "p1203")


def test_p1203_image(image_sample):
    from ayase.modules.p1203 import P1203Module

    m = P1203Module()
    m._ml_available = True
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.p1203_mos is None


def test_p1203_no_metadata(video_sample):
    from ayase.modules.p1203 import P1203Module

    m = P1203Module()
    m.setup()
    result = m.process(video_sample)
    assert result.quality_metrics is None or result.quality_metrics.p1203_mos is None


def test_p1203_with_metadata(synthetic_video):
    from ayase.models import VideoMetadata
    from ayase.modules.p1203 import P1203Module

    m = P1203Module()
    m.setup()
    sample = Sample(
        path=synthetic_video,
        is_video=True,
        video_metadata=VideoMetadata(
            width=1920,
            height=1080,
            frame_count=300,
            fps=30.0,
            duration=10.0,
            codec="h264",
            bitrate=5000000,
            file_size=6250000,
        ),
    )
    result = m.process(sample)
    assert result.quality_metrics is not None
    assert result.quality_metrics.p1203_mos is not None
    assert 1.0 <= result.quality_metrics.p1203_mos <= 5.0


def test_p1203_low_bitrate_lower_mos(synthetic_video):
    from ayase.models import VideoMetadata
    from ayase.modules.p1203 import P1203Module

    m = P1203Module()
    m.setup()
    high_br = Sample(
        path=synthetic_video,
        is_video=True,
        video_metadata=VideoMetadata(
            width=1920,
            height=1080,
            frame_count=300,
            fps=30.0,
            duration=10.0,
            codec="h264",
            bitrate=10000000,
            file_size=12500000,
        ),
    )
    low_br = Sample(
        path=synthetic_video,
        is_video=True,
        video_metadata=VideoMetadata(
            width=1920,
            height=1080,
            frame_count=300,
            fps=30.0,
            duration=10.0,
            codec="h264",
            bitrate=500000,
            file_size=625000,
        ),
    )
    r_high = m.process(high_br)
    r_low = m.process(low_br)
    assert r_high.quality_metrics.p1203_mos >= r_low.quality_metrics.p1203_mos


def test_industry_fields():
    qm = QualityMetrics()
    for field in ["cambi", "xpsnr", "vmaf_phone", "vmaf_4k"]:
        assert hasattr(qm, field)
    for field in ["visqol", "dnsmos_overall", "dnsmos_sig", "dnsmos_bak"]:
        assert hasattr(qm, field)
    for field in ["pu_psnr", "pu_ssim", "max_fall", "max_cll", "hdr_vdp", "delta_ictcp"]:
        assert hasattr(qm, field)
    for field in ["ciede2000", "psnr_hvs", "psnr_hvs_m", "cgvqm", "strred", "p1203_mos"]:
        assert hasattr(qm, field)


def test_industry_dataset_stats_fields():
    from ayase.models import DatasetStats

    ds = DatasetStats(total_samples=0, valid_samples=0, invalid_samples=0, total_size=0)
    for field in ["bd_rate", "bd_psnr"]:
        assert hasattr(ds, field)


def test_psnr_hvs_differs_from_plain_psnr(tmp_dir):
    """PSNR-HVS (CSF-weighted DCT) should produce different values than plain PSNR."""
    from ayase.modules.psnr_hvs import PSNRHVSModule

    # Create two similar images with known difference
    ref = np.full((64, 64, 3), 128, dtype=np.uint8)
    dist = ref.copy()
    dist[16:48, 16:48, :] = 100  # add a visible difference

    ref_path = tmp_dir / "ref.png"
    dist_path = tmp_dir / "dist.png"
    cv2.imwrite(str(ref_path), ref)
    cv2.imwrite(str(dist_path), dist)

    module = PSNRHVSModule()
    module.setup()
    hvs_score = module.compute_reference_score(dist_path, ref_path)

    # Compute plain PSNR for comparison
    mse = np.mean((ref.astype(float) - dist.astype(float)) ** 2)
    plain_psnr = 10.0 * np.log10(255.0 ** 2 / mse)

    assert hvs_score is not None
    # They should differ because HVS applies CSF weighting
    assert abs(hvs_score - plain_psnr) > 0.5, (
        f"PSNR-HVS ({hvs_score:.2f}) should differ from plain PSNR ({plain_psnr:.2f})"
    )


def test_p1203_dispatches_to_official_when_loaded():
    """P.1203 dispatches to official backend when backend=='official'."""
    from ayase.modules.p1203 import P1203Module

    module = P1203Module()
    module._backend = "official"
    module._ml_available = True

    mock_cls = MagicMock()
    mock_instance = MagicMock()
    mock_instance.calculate.return_value = {"O46": 3.5}
    mock_cls.return_value = mock_instance
    module._p1203_cls = mock_cls

    sample = Sample(
        path=Path("test.mp4"),
        is_video=True,
        video_metadata=VideoMetadata(
            width=1920, height=1080, frame_count=240,
            fps=24.0, duration=10.0, bitrate=5000000, file_size=6250000,
        ),
    )
    result = module.process(sample)
    assert result.quality_metrics is not None
    assert result.quality_metrics.p1203_mos == 3.5
    mock_cls.assert_called_once()



"""Tests for clipvqa module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_clipvqa_basics():
    from ayase.modules.clipvqa import CLIPVQAModule
    _test_module_basics(CLIPVQAModule, "clipvqa")

def test_clipvqa_image(image_sample):
    from ayase.modules.clipvqa import CLIPVQAModule
    image_sample.quality_metrics = QualityMetrics()
    m = CLIPVQAModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_clipvqa_video(video_sample):
    from ayase.modules.clipvqa import CLIPVQAModule
    video_sample.quality_metrics = QualityMetrics()
    m = CLIPVQAModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Jump cut detection module.

From Open-Sora 2.0 pipeline. Detects abrupt scene transitions
within a clip that indicate editing cuts.
"""

import logging
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class JumpCutModule(PipelineModule):
    name = "jump_cut"
    description = "Jump cut / abrupt transition detection (0-1, 1=no cuts)"
    default_config = {"threshold": 40.0}

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        if not sample.is_video:
            return sample

        try:
            import cv2

            cap = cv2.VideoCapture(str(sample.path))
            fps = cap.get(cv2.CAP_PROP_FPS) or 30.0
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            threshold = self.config.get("threshold", 40.0)

            # Sample densely for cut detection (every ~3 frames)
            step = max(1, int(fps / 10))
            prev_hist = None
            jump_count = 0
            frame_pairs = 0

            for i in range(0, total, step):
                cap.set(cv2.CAP_PROP_POS_FRAMES, i)
                ret, frame = cap.read()
                if not ret:
                    break

                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                hist = cv2.calcHist([gray], [0], None, [64], [0, 256])
                hist = hist.flatten() / (hist.sum() + 1e-8)

                if prev_hist is not None:
                    # Chi-squared distance between histograms
                    chi2 = float(np.sum((hist - prev_hist) ** 2 / (hist + prev_hist + 1e-8)))
                    if chi2 > threshold:
                        jump_count += 1
                    frame_pairs += 1

                prev_hist = hist
            cap.release()

            if frame_pairs == 0:
                sample.quality_metrics.jump_cut_score = 1.0
            else:
                duration = total / fps
                cuts_per_second = jump_count / max(duration, 0.1)
                sample.quality_metrics.jump_cut_score = 1.0 / (1.0 + cuts_per_second * 5.0)
        except Exception as e:
            logger.warning("Jump cut detection failed: %s", e)
        return sample



"""Tests for rankdvqa module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_rankdvqa_basics():
    from ayase.modules.rankdvqa import RankDVQAModule
    _test_module_basics(RankDVQAModule, "rankdvqa")

def test_rankdvqa_no_reference(image_sample):
    from ayase.modules.rankdvqa import RankDVQAModule
    m = RankDVQAModule()
    result = m.process(image_sample)
    assert result is image_sample



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

import torch.utils.data

from .coco import build as build_coco
from .objects365 import build as build_objects365
from .torchvision_datasets import CocoDetection


def get_coco_api_from_dataset(dataset):
    for _ in range(10):
        # if isinstance(dataset, torchvision.datasets.CocoDetection):
        #     break
        if isinstance(dataset, torch.utils.data.Subset):
            dataset = dataset.dataset
    if isinstance(dataset, CocoDetection):
        return dataset.coco


def build_dataset(image_set, args):
    if args.dataset_file == "objects365":
        return build_objects365(image_set, args)
    if args.dataset_file == "coco":
        return build_coco(image_set, args)
    if args.dataset_file == "coco_panoptic":
        # to avoid making panopticapi required for coco
        from .coco_panoptic import build as build_coco_panoptic

        return build_coco_panoptic(image_set, args)
    raise ValueError(f"dataset {args.dataset_file} not supported")



"""Tests for concept_presence module."""

from ayase.models import QualityMetrics
from ..conftest import _test_module_basics


def test_concept_presence_basics():
    from ayase.modules.concept_presence import ConceptPresenceModule

    _test_module_basics(ConceptPresenceModule, "concept_presence")


def test_concept_presence_skip_no_ml(image_sample):
    """Module should return sample unchanged when _ml_available is False."""
    from ayase.modules.concept_presence import ConceptPresenceModule

    image_sample.quality_metrics = QualityMetrics()
    m = ConceptPresenceModule()
    # Do NOT call on_mount() / setup(), so _ml_available stays False
    result = m.process(image_sample)
    assert result is image_sample
    assert result.quality_metrics.concept_presence is None
    assert result.quality_metrics.concept_count is None
    assert result.quality_metrics.concept_face_count is None


def test_concept_presence_image_with_mount(image_sample):
    """Module should process image sample after on_mount (test mode = heuristic)."""
    from ayase.modules.concept_presence import ConceptPresenceModule

    image_sample.quality_metrics = QualityMetrics()
    m = ConceptPresenceModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample


def test_concept_presence_video(video_sample):
    """Module should handle video sample gracefully."""
    from ayase.modules.concept_presence import ConceptPresenceModule

    video_sample.quality_metrics = QualityMetrics()
    m = ConceptPresenceModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample


def test_concept_presence_auto_mode():
    """Auto mode should detect face-related keywords."""
    from ayase.modules.concept_presence import ConceptPresenceModule

    m = ConceptPresenceModule()
    # Without any backends, auto_detect_mode falls back to "clip"
    assert m._auto_detect_mode(["a sunset over mountains"]) == "clip"
    assert m._auto_detect_mode(["a person smiling"]) == "clip"  # No face backend yet

    # Simulate having a face backend
    m._face_backend = "haar"
    mode = m._auto_detect_mode(["a person smiling"])
    assert mode in ("face", "combined")

    # Non-face concepts with face backend still returns "clip"
    assert m._auto_detect_mode(["a sunset over mountains"]) == "clip"


def test_concept_presence_fields():
    """Verify all concept_presence fields exist in QualityMetrics."""
    qm = QualityMetrics()
    assert hasattr(qm, "concept_presence")
    assert hasattr(qm, "concept_count")
    assert hasattr(qm, "concept_face_count")
    assert qm.concept_presence is None
    assert qm.concept_count is None
    assert qm.concept_face_count is None


def test_concept_presence_field_groups():
    """Verify field group assignments."""
    qm = QualityMetrics()
    groups = qm._FIELD_GROUPS
    assert groups.get("concept_presence") == "scene"
    assert groups.get("concept_count") == "scene"
    assert groups.get("concept_face_count") == "face"


def test_concept_presence_face_keywords():
    """Test that face keywords set is populated correctly."""
    from ayase.modules.concept_presence import _FACE_KEYWORDS

    assert "face" in _FACE_KEYWORDS
    assert "person" in _FACE_KEYWORDS
    assert "portrait" in _FACE_KEYWORDS
    assert "sunset" not in _FACE_KEYWORDS



"""Tests for hdr_metadata module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_hdr_metadata_basics():
    from ayase.modules.hdr_metadata import HDRMetadataModule
    _test_module_basics(HDRMetadataModule, "hdr_metadata")

def test_hdr_metadata_video(video_sample):
    from ayase.modules.hdr_metadata import HDRMetadataModule
    video_sample.quality_metrics = QualityMetrics()
    m = HDRMetadataModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Tests for viideo module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_viideo_basics():
    from ayase.modules.viideo import VIIDEOModule
    _test_module_basics(VIIDEOModule, "viideo")

def test_viideo_video(video_sample):
    from ayase.modules.viideo import VIIDEOModule
    video_sample.quality_metrics = QualityMetrics()
    m = VIIDEOModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------

import torch

def to_cuda(samples, targets, device):
    samples = samples.to(device, non_blocking=True)
    targets = [{k: v.to(device, non_blocking=True) for k, v in t.items()} for t in targets]
    return samples, targets

class data_prefetcher():
    def __init__(self, loader, device, prefetch=True):
        self.loader = iter(loader)
        self.prefetch = prefetch
        self.device = device
        if prefetch:
            self.stream = torch.cuda.Stream()
            self.preload()

    def preload(self):
        try:
            self.next_samples, self.next_targets = next(self.loader)
        except StopIteration:
            self.next_samples = None
            self.next_targets = None
            return
        # if record_stream() doesn't work, another option is to make sure device inputs are created
        # on the main stream.
        # self.next_input_gpu = torch.empty_like(self.next_input, device='cuda')
        # self.next_target_gpu = torch.empty_like(self.next_target, device='cuda')
        # Need to make sure the memory allocated for next_* is not still in use by the main stream
        # at the time we start copying to next_*:
        # self.stream.wait_stream(torch.cuda.current_stream())
        with torch.cuda.stream(self.stream):
            self.next_samples, self.next_targets = to_cuda(self.next_samples, self.next_targets, self.device)
            # more code for the alternative if record_stream() doesn't work:
            # copy_ will record the use of the pinned source tensor in this side stream.
            # self.next_input_gpu.copy_(self.next_input, non_blocking=True)
            # self.next_target_gpu.copy_(self.next_target, non_blocking=True)
            # self.next_input = self.next_input_gpu
            # self.next_target = self.next_target_gpu

            # With Amp, it isn't necessary to manually convert data to half.
            # if args.fp16:
            #     self.next_input = self.next_input.half()
            # else:

    def next(self):
        if self.prefetch:
            torch.cuda.current_stream().wait_stream(self.stream)
            samples = self.next_samples
            targets = self.next_targets
            if samples is not None:
                samples.record_stream(torch.cuda.current_stream())
            if targets is not None:
                for t in targets:
                    for k, v in t.items():
                        v.record_stream(torch.cuda.current_stream())
            self.preload()
        else:
            try:
                samples, targets = next(self.loader)
                samples, targets = to_cuda(samples, targets, self.device)
            except StopIteration:
                samples = None
                targets = None
        return samples, targets



"""FAVER — Blind Quality Prediction of Variable Frame Rate Videos.

Signal Processing 2024 — first NR-VQA targeting variable/high
frame rate distortions via bandpass temporal statistics + motion-aware
deep features.

faver_score — higher = better quality
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class FAVERModule(PipelineModule):
    name = "faver"
    description = "FAVER blind VQA for variable frame rate videos (2024)"
    default_config = {
        "subsample": 16,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 16)
        self._backend = "heuristic"

    def setup(self) -> None:
        try:
            import faver
            self._model = faver
            self._backend = "native"
            logger.info("FAVER (native) initialised")
            return
        except ImportError:
            pass

        self._backend = "heuristic"
        logger.info("FAVER (heuristic) — install faver for full model")

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample

        try:
            score = (
                float(self._model.predict(str(sample.path)))
                if self._backend == "native"
                else self._process_heuristic(sample)
            )

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.faver_score = score

        except Exception as e:
            logger.warning(f"FAVER failed for {sample.path}: {e}")

        return sample

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: bandpass temporal statistics + frame quality."""
        cap = cv2.VideoCapture(str(sample.path))
        fps = cap.get(cv2.CAP_PROP_FPS)
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        if total <= 1:
            cap.release()
            return None

        indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
        grays = []
        qualities = []

        for idx in indices:
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            if not ret:
                continue

            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            gray = cv2.resize(gray, (320, 240))
            grays.append(gray)

            # Per-frame spatial quality
            lap = cv2.Laplacian(gray, cv2.CV_64F).var()
            qualities.append(min(lap / 500.0, 1.0))

        cap.release()

        if len(grays) < 2:
            return None

        # Temporal bandpass statistics
        frame_diffs = []
        for i in range(len(grays) - 1):
            diff = np.mean(np.abs(grays[i + 1] - grays[i]))
            frame_diffs.append(diff)

        diff_arr = np.array(frame_diffs)

        # Bandpass: variance of differences (irregular = bad for VFR)
        temporal_regularity = 1.0 / (1.0 + np.var(diff_arr) * 0.01)

        # Motion-aware: mean difference level
        motion_level = np.mean(diff_arr)
        motion_quality = 1.0 / (1.0 + abs(motion_level - 5.0) * 0.05)

        # Frame rate factor
        fps_quality = min(fps / 30.0, 1.0) if fps > 0 else 0.5

        spatial_quality = float(np.mean(qualities))

        score = (
            0.35 * spatial_quality
            + 0.25 * temporal_regularity
            + 0.20 * motion_quality
            + 0.20 * fps_quality
        )

        return float(np.clip(score, 0.0, 1.0))



"""Multi-View Consistency module.

Estimates geometric consistency for video content using feature
matching across frames:

  multiview_consistency — 0-1 (higher = better 3D consistency)

Algorithm:
  1. Extract keypoints (ORB) from frame pairs.
  2. Match features and compute fundamental matrix via RANSAC.
  3. Measure inlier ratio — high inlier ratio = geometrically
     consistent views.
  4. Check epipolar constraint residuals.

This is a proxy for 3D geometric consistency without requiring
multi-view input — it measures whether frame-to-frame geometry
is coherent enough to support 3D reconstruction.
"""

import logging
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class MultiViewConsistencyModule(PipelineModule):
    name = "multi_view_consistency"
    description = "Geometric multi-view consistency via epipolar analysis"
    default_config = {
        "subsample": 5,  # Frame gap for pair analysis
        "max_pairs": 30,
        "min_matches": 20,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 5)
        self.max_pairs = self.config.get("max_pairs", 30)
        self.min_matches = self.config.get("min_matches", 20)

        self._orb = None

    def setup(self) -> None:
        self._orb = cv2.ORB_create(nfeatures=500)
        self._bf = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
        logger.info("Multi-view consistency: ORB + RANSAC initialised")

    def _compute_pair_consistency(
        self, gray1: np.ndarray, gray2: np.ndarray
    ) -> Optional[float]:
        """Compute geometric consistency between two frames.

        Returns inlier ratio (0-1) from fundamental matrix estimation.
        """
        kp1, des1 = self._orb.detectAndCompute(gray1, None)
        kp2, des2 = self._orb.detectAndCompute(gray2, None)

        if des1 is None or des2 is None:
            return None
        if len(kp1) < self.min_matches or len(kp2) < self.min_matches:
            return None

        matches = self._bf.match(des1, des2)
        if len(matches) < self.min_matches:
            return None

        # Sort by distance
        matches = sorted(matches, key=lambda m: m.distance)

        pts1 = np.float32([kp1[m.queryIdx].pt for m in matches])
        pts2 = np.float32([kp2[m.trainIdx].pt for m in matches])

        # Estimate fundamental matrix via RANSAC
        F, mask = cv2.findFundamentalMat(
            pts1, pts2, cv2.FM_RANSAC, ransacReprojThreshold=3.0
        )

        if mask is None:
            return None

        inlier_ratio = float(mask.sum()) / max(len(mask), 1)
        return inlier_ratio

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video or self._orb is None:
            return sample

        try:
            score = self._process_video(sample.path)
            if score is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.multiview_consistency = score

            logger.debug(
                f"Multi-view consistency for {sample.path.name}: {score:.3f}"
            )

        except Exception as e:
            logger.error(f"Multi-view consistency failed for {sample.path}: {e}")

        return sample

    def _process_video(self, path: Path) -> Optional[float]:
        cap = cv2.VideoCapture(str(path))
        if not cap.isOpened():
            return None

        prev_gray = None
        consistencies = []
        idx = 0
        pairs = 0

        while pairs < self.max_pairs:
            ret, frame = cap.read()
            if not ret:
                break

            if idx % self.subsample == 0:
                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

                if prev_gray is not None:
                    c = self._compute_pair_consistency(prev_gray, gray)
                    if c is not None:
                        consistencies.append(c)
                        pairs += 1

                prev_gray = gray

            idx += 1

        cap.release()

        if not consistencies:
            return None

        return float(np.mean(consistencies))



# --------------------------------------------------------
# BEATs: Audio Pre-Training with Acoustic Tokenizers (https://arxiv.org/abs/2212.09058)
# Github source: https://github.com/microsoft/unilm/tree/master/beats
# Copyright (c) 2022 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import math
import warnings
import torch
from torch import Tensor, nn
import torch.nn.functional as F


class GradMultiply(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale):
        ctx.scale = scale
        res = x.new(x)
        return res

    @staticmethod
    def backward(ctx, grad):
        return grad * ctx.scale, None


class SamePad(nn.Module):
    def __init__(self, kernel_size, causal=False):
        super().__init__()
        if causal:
            self.remove = kernel_size - 1
        else:
            self.remove = 1 if kernel_size % 2 == 0 else 0

    def forward(self, x):
        if self.remove > 0:
            x = x[:, :, : -self.remove]
        return x


class Swish(nn.Module):
    def __init__(self):
        super(Swish, self).__init__()
        self.act = torch.nn.Sigmoid()

    def forward(self, x):
        return x * self.act(x)


class GLU_Linear(nn.Module):
    def __init__(self, input_dim, output_dim, glu_type="sigmoid", bias_in_glu=True):
        super(GLU_Linear, self).__init__()

        self.glu_type = glu_type
        self.output_dim = output_dim

        if glu_type == "sigmoid":
            self.glu_act = torch.nn.Sigmoid()
        elif glu_type == "swish":
            self.glu_act = Swish()
        elif glu_type == "relu":
            self.glu_act = torch.nn.ReLU()
        elif glu_type == "gelu":
            self.glu_act = torch.nn.GELU()

        if bias_in_glu:
            self.linear = nn.Linear(input_dim, output_dim * 2, True)
        else:
            self.linear = nn.Linear(input_dim, output_dim * 2, False)

    def forward(self, x):
        # to be consistent with GLU_Linear, we assume the input always has the #channel (#dim) in the last dimension of the tensor, so need to switch the dimension first for 1D-Conv case
        x = self.linear(x)

        if self.glu_type == "bilinear":
            x = (x[:, :, 0:self.output_dim] * x[:, :, self.output_dim:self.output_dim * 2])
        else:
            x = (x[:, :, 0:self.output_dim] * self.glu_act(x[:, :, self.output_dim:self.output_dim * 2]))

        return x


def gelu_accurate(x):
    if not hasattr(gelu_accurate, "_a"):
        gelu_accurate._a = math.sqrt(2 / math.pi)
    return (
        0.5 * x * (1 + torch.tanh(gelu_accurate._a * (x + 0.044715 * torch.pow(x, 3))))
    )


def gelu(x: torch.Tensor) -> torch.Tensor:
    return torch.nn.functional.gelu(x.float()).type_as(x)


def get_activation_fn(activation: str):
    """Returns the activation function corresponding to `activation`"""

    if activation == "relu":
        return F.relu
    elif activation == "gelu":
        return gelu
    elif activation == "gelu_fast":
        warnings.warn(
            "--activation-fn=gelu_fast has been renamed to gelu_accurate"
        )
        return gelu_accurate
    elif activation == "gelu_accurate":
        return gelu_accurate
    elif activation == "tanh":
        return torch.tanh
    elif activation == "linear":
        return lambda x: x
    elif activation == "glu":
        return lambda x: x
    else:
        raise RuntimeError("--activation-fn {} not supported".format(activation))


def quant_noise(module, p, block_size):
    """
    Wraps modules and applies quantization noise to the weights for
    subsequent quantization with Iterative Product Quantization as
    described in "Training with Quantization Noise for Extreme Model Compression"

    Args:
        - module: nn.Module
        - p: amount of Quantization Noise
        - block_size: size of the blocks for subsequent quantization with iPQ

    Remarks:
        - Module weights must have the right sizes wrt the block size
        - Only Linear, Embedding and Conv2d modules are supported for the moment
        - For more detail on how to quantize by blocks with convolutional weights,
          see "And the Bit Goes Down: Revisiting the Quantization of Neural Networks"
        - We implement the simplest form of noise here as stated in the paper
          which consists in randomly dropping blocks
    """

    # if no quantization noise, don't register hook
    if p <= 0:
        return module

    # supported modules
    assert isinstance(module, (nn.Linear, nn.Embedding, nn.Conv2d))

    # test whether module.weight has the right sizes wrt block_size
    is_conv = module.weight.ndim == 4

    # 2D matrix
    if not is_conv:
        assert (
            module.weight.size(1) % block_size == 0
        ), "Input features must be a multiple of block sizes"

    # 4D matrix
    else:
        # 1x1 convolutions
        if module.kernel_size == (1, 1):
            assert (
                module.in_channels % block_size == 0
            ), "Input channels must be a multiple of block sizes"
        # regular convolutions
        else:
            k = module.kernel_size[0] * module.kernel_size[1]
            assert k % block_size == 0, "Kernel size must be a multiple of block size"

    def _forward_pre_hook(mod, input):
        # no noise for evaluation
        if mod.training:
            if not is_conv:
                # gather weight and sizes
                weight = mod.weight
                in_features = weight.size(1)
                out_features = weight.size(0)

                # split weight matrix into blocks and randomly drop selected blocks
                mask = torch.zeros(
                    in_features // block_size * out_features, device=weight.device
                )
                mask.bernoulli_(p)
                mask = mask.repeat_interleave(block_size, -1).view(-1, in_features)

            else:
                # gather weight and sizes
                weight = mod.weight
                in_channels = mod.in_channels
                out_channels = mod.out_channels

                # split weight matrix into blocks and randomly drop selected blocks
                if mod.kernel_size == (1, 1):
                    mask = torch.zeros(
                        int(in_channels // block_size * out_channels),
                        device=weight.device,
                    )
                    mask.bernoulli_(p)
                    mask = mask.repeat_interleave(block_size, -1).view(-1, in_channels)
                else:
                    mask = torch.zeros(
                        weight.size(0), weight.size(1), device=weight.device
                    )
                    mask.bernoulli_(p)
                    mask = (
                        mask.unsqueeze(2)
                        .unsqueeze(3)
                        .repeat(1, 1, mod.kernel_size[0], mod.kernel_size[1])
                    )

            # scale weights and apply mask
            mask = mask.to(
                torch.bool
            )  # x.bool() is not currently supported in TorchScript
            s = 1 / (1 - p)
            mod.weight.data = s * weight.masked_fill(mask, 0)

    module.register_forward_pre_hook(_forward_pre_hook)
    return module




"""DISTS (Deep Image Structure and Texture Similarity) module.

DISTS is a full-reference perceptual similarity metric that explicitly
separates structure and texture.  It correlates better with human
perception than SSIM/MS-SSIM for texture-rich content.

Range: 0-1 (lower = more similar / better quality).
Requires a reference video or image.

Uses the ``piq`` package (already an Ayase dependency).
"""

import logging
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class DISTSModule(PipelineModule):
    name = "dists"
    description = "Deep Image Structure and Texture Similarity (full-reference)"
    default_config = {
        "subsample": 5,  # Process every Nth video frame
        "warning_threshold": 0.3,  # Warn if DISTS > 0.3 (lower is better)
        "device": "auto",
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 5)
        self.warning_threshold = self.config.get("warning_threshold", 0.3)
        self.device_config = self.config.get("device", "auto")
        self.device = None
        self._ml_available = False
        self._dists_fn = None

    def setup(self) -> None:
        try:
            import torch
            from piq import DISTS as PIQ_DISTS

            if self.device_config == "auto":
                self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            else:
                self.device = torch.device(self.device_config)

            self._dists_fn = PIQ_DISTS().to(self.device)
            self._ml_available = True
            logger.info(f"DISTS module initialised on {self.device}")

        except ImportError:
            logger.warning("piq not installed. Install with: pip install piq")
        except Exception as e:
            logger.warning(f"Failed to setup DISTS: {e}")

    # ------------------------------------------------------------------
    # helpers
    # ------------------------------------------------------------------

    def _to_tensor(self, img_bgr: np.ndarray):
        """BGR uint8 (H,W,C) → float32 (1,C,H,W) in [0,1]."""
        import torch

        rgb = cv2.cvtColor(img_bgr, cv2.COLOR_BGR2RGB)
        t = torch.from_numpy(rgb).permute(2, 0, 1).unsqueeze(0).float() / 255.0
        return t.to(self.device)

    def _score_pair(self, ref: np.ndarray, dist: np.ndarray) -> Optional[float]:
        """Compute DISTS for one frame-pair."""
        try:
            import torch

            ref_t = self._to_tensor(ref)
            dist_t = self._to_tensor(dist)

            # DISTS requires both inputs to have the same spatial size
            if ref_t.shape != dist_t.shape:
                # Resize distorted to reference size
                h, w = ref_t.shape[2], ref_t.shape[3]
                dist_t = torch.nn.functional.interpolate(
                    dist_t, size=(h, w), mode="bilinear", align_corners=False
                )

            with torch.no_grad():
                score = self._dists_fn(dist_t, ref_t)
            return float(score.item())
        except Exception as e:
            logger.debug(f"DISTS pair computation failed: {e}")
            return None

    # ------------------------------------------------------------------
    # process
    # ------------------------------------------------------------------

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available:
            return sample

        reference = getattr(sample, "reference_path", None)
        if reference is None:
            return sample
        reference = Path(reference) if not isinstance(reference, Path) else reference
        if not reference.exists():
            return sample

        try:
            if sample.is_video:
                score = self._process_video(sample.path, reference)
            else:
                score = self._process_image(sample.path, reference)

            if score is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.dists = score

            if score > self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"High DISTS (structure/texture dissimilarity): {score:.3f}",
                        details={"dists": score, "threshold": self.warning_threshold},
                        recommendation=(
                            "Content diverges from reference in structure or "
                            "texture. Inspect for hallucinated details or "
                            "texture loss."
                        ),
                    )
                )

            logger.debug(f"DISTS for {sample.path.name}: {score:.4f}")

        except Exception as e:
            logger.error(f"DISTS failed for {sample.path}: {e}")

        return sample

    def _process_image(self, path: Path, ref_path: Path) -> Optional[float]:
        ref = cv2.imread(str(ref_path))
        dist = cv2.imread(str(path))
        if ref is None or dist is None:
            return None
        return self._score_pair(ref, dist)

    def _process_video(self, path: Path, ref_path: Path) -> Optional[float]:
        ref_cap = cv2.VideoCapture(str(ref_path))
        dist_cap = cv2.VideoCapture(str(path))
        scores = []
        idx = 0

        while True:
            r1, ref_frame = ref_cap.read()
            r2, dist_frame = dist_cap.read()
            if not r1 or not r2:
                break
            if idx % self.subsample == 0:
                s = self._score_pair(ref_frame, dist_frame)
                if s is not None:
                    scores.append(s)
            idx += 1

        ref_cap.release()
        dist_cap.release()
        return float(np.mean(scores)) if scores else None



import torch
from .clip import clip
from torchvision import transforms
import numpy as np
import cv2
from PIL import Image


class CLIPFeatureExtractor():
    def __init__(self, device=None):
        """
        Initialize the CLIP Image feature extractor.

        Parameters:
        - device (torch.device): Device to run the model on. Defaults to CUDA if available.
        """
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.model, self.preprocess = self._load_model()
        
    def _load_model(self):
        """
        Load the CLIP model.

        Parameters:

        Returns:
        - model (torch.nn.Module): Loaded CLIP model.
        """
        model, preprocess = clip.load("ViT-B/32", device=self.device)
        from .clipiqa_arch import CLIPIQA 
        model = CLIPIQA()
        return model, preprocess
    
    def extract_features(self, frames, batch_size=80):
        """
        Extract features from a single frame or a batch of frames.

        :param frames: A single frame (H x W x C) or a batch of frames (N x H x W x C).
        :return: Extracted features as a NumPy array.
        """
        frames = np.array(frames)
        if isinstance(frames, np.ndarray):
            if frames.ndim == 3:
                frames = np.expand_dims(frames, axis=0)  # Convert single frame to batch
        else:
            raise ValueError("Input frames should be a list og NumPy array")
        
        # print number of frames 
        #print("Number of frames: ", frames.shape[0])    
        
        feats = []
        with torch.no_grad():
            # batch wise 
            for i in range(0, frames.shape[0], batch_size):
                # Get the features from the CLIP model
                images = frames[i:i+batch_size]
                images = [
                    self.preprocess(
                        Image.fromarray(
                            np.clip(image * 255.0, 0, 255).astype(np.uint8),
                            "RGB",
                        )
                    )
                    for image in images
                ]
                images = torch.stack(images).to(self.device)
                #feat = self.model.encode_image(images)
                feat = self.model(images)
                # extend the feats list 
                feats.extend(feat.cpu().numpy())
            # normalize the features
            feats = np.array(feats)
            feats /= np.linalg.norm(feats, axis=-1, keepdims=True)
        return feats
        



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import datetime
import logging
import os
import time

from ...common import dist_utils
import torch
import torch.distributed as dist
import torch.nn.functional as F
from ...common.dist_utils import download_cached_file
from ...common.logger import MetricLogger
from ...common.utils import is_url
from ...models.base_model import BaseModel
from ...models.vit import interpolate_pos_embed
from transformers import BertTokenizer


class AlbefBase(BaseModel):
    @classmethod
    def init_tokenizer(cls):
        return BertTokenizer.from_pretrained("bert-base-uncased")

    def load_from_pretrained(self, url_or_filename, rename_text_keys=True):
        if is_url(url_or_filename):
            cached_file = download_cached_file(
                url_or_filename, check_hash=False, progress=True
            )
            checkpoint = torch.load(cached_file, map_location="cpu")
        elif os.path.isfile(url_or_filename):
            checkpoint = torch.load(url_or_filename, map_location="cpu")
        else:
            raise RuntimeError("checkpoint url or path is invalid")

        if "model" in checkpoint:
            state_dict = checkpoint["model"]
        else:
            state_dict = checkpoint

        state_dict["visual_encoder.pos_embed"] = interpolate_pos_embed(
            state_dict["visual_encoder.pos_embed"], self.visual_encoder
        )
        if (
            "visual_encoder_m.pos_embed" in self.state_dict().keys()
            and "visual_encoder_m.pos_embed" in state_dict
        ):
            state_dict["visual_encoder_m.pos_embed"] = interpolate_pos_embed(
                state_dict["visual_encoder_m.pos_embed"], self.visual_encoder_m
            )

        if rename_text_keys:
            for key in list(state_dict.keys()):
                if "bert" in key:
                    new_key = key.replace("bert.", "")
                    state_dict[new_key] = state_dict[key]
                    del state_dict[key]

        for key in self.state_dict().keys():
            if key in state_dict.keys():
                if state_dict[key].shape != self.state_dict()[key].shape:
                    del state_dict[key]

        msg = self.load_state_dict(state_dict, strict=False)

        logging.info("Missing keys {}".format(msg.missing_keys))
        logging.info("load checkpoint from %s" % url_or_filename)
        return msg


def compute_sim_matrix(model, data_loader, **kwargs):
    k_test = kwargs.pop("k_test")

    metric_logger = MetricLogger(delimiter="  ")
    header = "Evaluation:"

    logging.info("Computing features for evaluation...")
    start_time = time.time()

    texts = data_loader.dataset.text
    num_text = len(texts)
    text_bs = 256
    text_ids = []
    text_embeds = []
    text_atts = []
    for i in range(0, num_text, text_bs):
        text = texts[i : min(num_text, i + text_bs)]
        text_input = model.tokenizer(
            text,
            padding="max_length",
            truncation=True,
            max_length=35,
            return_tensors="pt",
        ).to(model.device)
        text_output = model.text_encoder.forward_text(text_input)
        text_embed = F.normalize(
            model.text_proj(text_output.last_hidden_state[:, 0, :])
        )
        text_embeds.append(text_embed)
        text_ids.append(text_input.input_ids)
        text_atts.append(text_input.attention_mask)

    text_embeds = torch.cat(text_embeds, dim=0)
    text_ids = torch.cat(text_ids, dim=0)
    text_atts = torch.cat(text_atts, dim=0)
    if hasattr(model.tokenizer, "enc_token_id"):
        text_ids[:, 0] = model.tokenizer.enc_token_id

    image_feats = []
    image_embeds = []
    for samples in data_loader:
        image = samples["image"]

        image = image.to(model.device)
        image_feat = model.visual_encoder.forward_features(image)
        image_embed = model.vision_proj(image_feat[:, 0, :])
        image_embed = F.normalize(image_embed, dim=-1)

        image_feats.append(image_feat.cpu())
        image_embeds.append(image_embed)

    image_feats = torch.cat(image_feats, dim=0)
    image_embeds = torch.cat(image_embeds, dim=0)

    sims_matrix = image_embeds @ text_embeds.t()
    score_matrix_i2t = torch.full(
        (len(data_loader.dataset.image), len(texts)), -100.0
    ).to(model.device)

    num_tasks = dist_utils.get_world_size()
    rank = dist_utils.get_rank()
    step = sims_matrix.size(0) // num_tasks + 1
    start = rank * step
    end = min(sims_matrix.size(0), start + step)

    for i, sims in enumerate(
        metric_logger.log_every(sims_matrix[start:end], 50, header)
    ):
        # topk_sim, topk_idx = sims.topk(k=config["k_test"], dim=0)
        topk_sim, topk_idx = sims.topk(k=k_test, dim=0)

        encoder_output = image_feats[start + i].repeat(k_test, 1, 1).to(model.device)
        encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(
            model.device
        )
        output = model.text_encoder(
            text_ids[topk_idx],
            attention_mask=text_atts[topk_idx],
            encoder_hidden_states=encoder_output,
            encoder_attention_mask=encoder_att,
            return_dict=True,
        )
        score = model.itm_head(output.last_hidden_state[:, 0, :])[:, 1]
        score_matrix_i2t[start + i, topk_idx] = score + topk_sim

    sims_matrix = sims_matrix.t()
    score_matrix_t2i = torch.full(
        (len(texts), len(data_loader.dataset.image)), -100.0
    ).to(model.device)

    step = sims_matrix.size(0) // num_tasks + 1
    start = rank * step
    end = min(sims_matrix.size(0), start + step)

    for i, sims in enumerate(
        metric_logger.log_every(sims_matrix[start:end], 50, header)
    ):

        topk_sim, topk_idx = sims.topk(k=k_test, dim=0)
        encoder_output = image_feats[topk_idx.cpu()].to(model.device)
        encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(
            model.device
        )
        output = model.text_encoder(
            text_ids[start + i].repeat(k_test, 1),
            attention_mask=text_atts[start + i].repeat(k_test, 1),
            encoder_hidden_states=encoder_output,
            encoder_attention_mask=encoder_att,
            return_dict=True,
        )
        score = model.itm_head(output.last_hidden_state[:, 0, :])[:, 1]
        score_matrix_t2i[start + i, topk_idx] = score + topk_sim

    if dist_utils.is_dist_avail_and_initialized():
        dist.barrier()
        torch.distributed.all_reduce(
            score_matrix_i2t, op=torch.distributed.ReduceOp.SUM
        )
        torch.distributed.all_reduce(
            score_matrix_t2i, op=torch.distributed.ReduceOp.SUM
        )

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    logging.info("Evaluation time {}".format(total_time_str))

    return score_matrix_i2t.cpu().numpy(), score_matrix_t2i.cpu().numpy()



"""Tests for BrightRate module."""

from ayase.models import QualityMetrics

from ..conftest import _test_module_basics


def test_brightrate_basics():
    from ayase.modules.brightrate import BrightRateModule

    _test_module_basics(BrightRateModule, "brightrate")


def test_brightrate_skip_without_backend(video_sample):
    from ayase.modules.brightrate import BrightRateModule

    video_sample.quality_metrics = QualityMetrics()
    result = BrightRateModule().process(video_sample)
    assert result is video_sample
    assert result.quality_metrics.brightrate_score is None


def test_brightrate_field_exists():
    qm = QualityMetrics()
    assert hasattr(qm, "brightrate_score")
    assert qm.brightrate_score is None


def test_brightrate_field_group():
    qm = QualityMetrics()
    assert qm._FIELD_GROUPS.get("brightrate_score") == "hdr"



"""Tests for ciede2000 module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_ciede2000_basics():
    from ayase.modules.ciede2000 import CIEDE2000Module
    _test_module_basics(CIEDE2000Module, "ciede2000")

def test_ciede2000_no_reference(image_sample):
    from ayase.modules.ciede2000 import CIEDE2000Module
    m = CIEDE2000Module()
    result = m.process(image_sample)
    assert result is image_sample



"""Tests for psnr_div module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_psnr_div_basics():
    from ayase.modules.psnr_div import PSNRDIVModule
    _test_module_basics(PSNRDIVModule, "psnr_div")

def test_psnr_div_no_reference(image_sample):
    from ayase.modules.psnr_div import PSNRDIVModule
    m = PSNRDIVModule()
    result = m.process(image_sample)
    assert result is image_sample



"""CLiF-VQA — Human Feelings VQA.

2024 — extracts human-feelings features from CLIP to simulate
HVS perceptual process for quality assessment.

clifvqa_score — higher = better quality
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class CLiFVQAModule(PipelineModule):
    name = "clifvqa"
    description = "CLiF-VQA human feelings VQA via CLIP (2024)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._backend = "heuristic"

    def setup(self) -> None:
        try:
            import clifvqa
            self._model = clifvqa
            self._backend = "native"
            logger.info("CLiF-VQA (native) initialised")
            return
        except ImportError:
            pass
        self._backend = "heuristic"
        logger.info("CLiF-VQA (heuristic)")

    def process(self, sample: Sample) -> Sample:
        try:
            score = (
                float(self._model.predict(str(sample.path)))
                if self._backend == "native"
                else self._process_heuristic(sample)
            )
            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.clifvqa_score = score
        except Exception as e:
            logger.warning(f"CLiF-VQA failed for {sample.path}: {e}")
        return sample

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: human feeling simulation via perceptual features."""
        frames = self._extract_frames(sample)
        if not frames:
            return None

        feeling_scores = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)

            # Comfort feeling: brightness + contrast balance
            brightness = gray.mean() / 255.0
            brightness_comfort = 1.0 - abs(brightness - 0.5) * 2.0

            # Clarity feeling: sharpness
            clarity = min(cv2.Laplacian(gray, cv2.CV_64F).var() / 500.0, 1.0)

            # Warmth feeling: color temperature proxy
            b, g, r = frame[:, :, 0].mean(), frame[:, :, 1].mean(), frame[:, :, 2].mean()
            warmth = min(r / (b + 1e-8), 2.0) / 2.0

            # Vibrancy feeling: saturation
            vibrancy = min(hsv[:, :, 1].astype(float).mean() / 128.0, 1.0)

            # Harmony feeling: color consistency (low std in hue)
            hue = hsv[:, :, 0].astype(float)
            harmony = 1.0 / (1.0 + hue.std() / 30.0)

            feeling = (
                0.25 * brightness_comfort
                + 0.30 * clarity
                + 0.15 * warmth
                + 0.15 * vibrancy
                + 0.15 * harmony
            )
            feeling_scores.append(feeling)

        return float(np.clip(np.mean(feeling_scores), 0.0, 1.0))

    def _extract_frames(self, sample: Sample):
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return []
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)
        return frames



"""STREAM — Spatio-Temporal Evaluation and Analysis Metric (ICLR 2024).

GitHub: https://github.com/pro2nit/STREAM
Distribution-level: stream_spatial, stream_temporal
"""
import logging
import cv2
import numpy as np
from typing import List, Optional

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import BatchMetricModule

logger = logging.getLogger(__name__)


class STREAMModule(BatchMetricModule):
    name = "stream_metric"
    description = "STREAM spatial/temporal generation eval (ICLR 2024)"
    default_config = {"subsample": 8}

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._temporal_features: List = []

    def extract_features(self, sample: Sample) -> Optional[object]:
        """Extract spatial + temporal features from a sample."""
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                try:
                    total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                    if total <= 1:
                        return None
                    indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
                    grays = []
                    for idx in indices:
                        cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                        ret, f = cap.read()
                        if ret:
                            grays.append(cv2.resize(
                                cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(float), (64, 64)
                            ))
                finally:
                    cap.release()

                if len(grays) < 2:
                    return None

                # Spatial feature: flattened mean frame
                spatial = np.mean([g.flatten() for g in grays], axis=0)

                # Temporal feature: FFT of frame diffs
                diffs = [np.mean(np.abs(grays[i] - grays[i + 1])) for i in range(len(grays) - 1)]
                fft_mag = np.abs(np.fft.fft(diffs))
                self._temporal_features.append(fft_mag[: len(fft_mag) // 2 + 1])

                return spatial
            else:
                img = cv2.imread(str(sample.path))
                if img is not None:
                    gray = cv2.resize(
                        cv2.cvtColor(img, cv2.COLOR_BGR2GRAY).astype(float), (64, 64)
                    )
                    return gray.flatten()
                return None
        except Exception as e:
            logger.warning(f"STREAM feature extraction failed: {e}")
            return None

    def compute_distribution_metric(
        self, features: List, reference_features: Optional[List] = None
    ) -> float:
        """Compute STREAM spatial and temporal scores."""
        spatial_score = 0.0
        temporal_score = 0.0

        if len(features) >= 2:
            feats = np.array(features)
            mean_feat = feats.mean(axis=0)
            diversity = np.mean(np.std(feats, axis=0))
            fidelity = np.mean(np.linalg.norm(feats - mean_feat, axis=1))
            spatial_score = float(diversity / (fidelity + 1e-8))

        if len(self._temporal_features) >= 2:
            max_len = max(len(t) for t in self._temporal_features)
            padded = [np.pad(t, (0, max_len - len(t))) for t in self._temporal_features]
            t_feats = np.array(padded)
            temporal_score = float(1.0 / (1.0 + np.var(t_feats.mean(axis=1))))

        # Store in pipeline stats if available
        if hasattr(self, "pipeline") and self.pipeline and hasattr(self.pipeline, "stats"):
            self.pipeline.stats["stream_spatial"] = spatial_score
            self.pipeline.stats["stream_temporal"] = temporal_score

        return spatial_score  # Return primary metric



"""UIQM — Underwater Image Quality Measure.

Panetta et al. 2016 — pure computation metric for underwater images.
Combines colorfulness (UICM), sharpness (UISM), and contrast (UIConM).

GitHub: https://github.com/tkrahn108/UIQM

uiqm_score — higher = better quality

Formula: UIQM = c1*UICM + c2*UISM + c3*UIConM
Default weights: c1=0.0282, c2=0.2953, c3=3.5753
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


def _uicm(img: np.ndarray) -> float:
    """Underwater Image Colorfulness Measure (UICM)."""
    r, g, b = img[:, :, 2].astype(float), img[:, :, 1].astype(float), img[:, :, 0].astype(float)
    rg = r - g
    yb = 0.5 * (r + g) - b

    mu_rg = np.mean(rg)
    mu_yb = np.mean(yb)
    sigma_rg = np.sqrt(np.var(rg))
    sigma_yb = np.sqrt(np.var(yb))

    return -0.0268 * np.sqrt(mu_rg ** 2 + mu_yb ** 2) + 0.1586 * np.sqrt(sigma_rg ** 2 + sigma_yb ** 2)


def _uism(img: np.ndarray) -> float:
    """Underwater Image Sharpness Measure (UISM) via Sobel edge detector."""
    # Compute edge maps per channel using Sobel
    channels = cv2.split(img)
    eme_vals = []

    for ch in channels:
        ch = ch.astype(np.float64)
        sobel_x = cv2.Sobel(ch, cv2.CV_64F, 1, 0, ksize=3)
        sobel_y = cv2.Sobel(ch, cv2.CV_64F, 0, 1, ksize=3)
        edge_map = np.sqrt(sobel_x ** 2 + sobel_y ** 2)

        # Block-based EME (Measure of Enhancement)
        h, w = edge_map.shape
        block_h, block_w = max(h // 8, 1), max(w // 8, 1)
        eme = 0.0
        count = 0
        for i in range(0, h - block_h + 1, block_h):
            for j in range(0, w - block_w + 1, block_w):
                block = edge_map[i : i + block_h, j : j + block_w]
                bmax = block.max()
                bmin = block.min()
                if bmin > 0 and bmax > 0:
                    eme += 20.0 * np.log(bmax / bmin + 1e-8)
                    count += 1

        eme_vals.append(eme / max(count, 1))

    # Weighted sum (lambda_r=0.299, lambda_g=0.587, lambda_b=0.114)
    return 0.299 * eme_vals[2] + 0.587 * eme_vals[1] + 0.114 * eme_vals[0]


def _uiconm(img: np.ndarray) -> float:
    """Underwater Image Contrast Measure (UIConM) via logAMEE."""
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY).astype(np.float64)
    h, w = gray.shape
    block_h, block_w = max(h // 8, 1), max(w // 8, 1)

    logamee = 0.0
    count = 0

    for i in range(0, h - block_h + 1, block_h):
        for j in range(0, w - block_w + 1, block_w):
            block = gray[i : i + block_h, j : j + block_w]
            bmax = block.max()
            bmin = block.min()
            if bmin > 0 and bmax > bmin:
                alpha = (bmax + bmin) / 2.0
                plip_add = bmax + bmin - bmax * bmin / alpha if alpha > 0 else 0
                plip_sub = abs(bmax - bmin)
                if plip_add > 0 and plip_sub > 0:
                    logamee += np.log(plip_sub / plip_add + 1e-8)
                    count += 1

    return logamee / max(count, 1)


class UIQMModule(PipelineModule):
    name = "uiqm"
    description = "UIQM underwater image quality measure (Panetta et al. 2016)"
    default_config = {
        "c1": 0.0282,
        "c2": 0.2953,
        "c3": 3.5753,
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.c1 = self.config.get("c1", 0.0282)
        self.c2 = self.config.get("c2", 0.2953)
        self.c3 = self.config.get("c3", 3.5753)
        self.subsample = self.config.get("subsample", 8)

    def process(self, sample: Sample) -> Sample:
        try:
            score = self._compute(sample)
            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.uiqm_score = score
        except Exception as e:
            logger.warning(f"UIQM failed for {sample.path}: {e}")
        return sample

    def _score_frame(self, frame: np.ndarray) -> float:
        uicm = _uicm(frame)
        uism = _uism(frame)
        uiconm = _uiconm(frame)
        return self.c1 * uicm + self.c2 * uism + self.c3 * uiconm

    def _compute(self, sample: Sample) -> Optional[float]:
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return None
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            scores = []
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    scores.append(self._score_frame(frame))
            cap.release()
            return float(np.mean(scores)) if scores else None
        else:
            img = cv2.imread(str(sample.path))
            if img is None:
                return None
            return float(self._score_frame(img))



from typing import List
import os
from PIL import Image, ImageSequence
import decord

VALID_DATA_FORMAT_STRING = "Input data must be {'.jpg', '.jpeg', '.png', '.tif'} for image; or {'.mp4', '.avi', '.webm', '.mov', '.mkv', '.wmv', '.gif'}  for videos!"

# 均匀抽帧，必采样首尾帧。
def sample_frame_indices(start_frame, total_frames: int, n_frames: int):
    if n_frames == 1:
        return [0]  # sample first frame in default
    sample_ids = [round(i * (total_frames - 1) / (n_frames - 1)) for i in range(n_frames)]
    sample_ids = [i + start_frame for i in sample_ids]
    return sample_ids

def sample_video(
    video_path: str, 
    n_frames: int = None,
    start_time: int = 0,
    end_time: int = -1
    ) -> List[Image.Image]:

    assert os.path.exists(video_path), f"File not found: {video_path}"
    vr = decord.VideoReader(video_path, num_threads=1, ctx=decord.cpu(0))
    vr.seek(0)
    total_frames = len(vr)
    fps = vr.get_avg_fps()

    start_frame = 0
    end_frame = total_frames - 1
    if start_time > 0:
        start_frame = min((total_frames-1), int(fps*start_time))
    if end_time > 0:
        end_frame = max(start_frame, int(fps*end_time))
        end_frame = min(end_frame, (total_frames-1))
    frame_indices = sample_frame_indices(
        start_frame=start_frame,
        total_frames=end_frame - start_frame + 1,
        n_frames=n_frames,
    )

    frames = vr.get_batch(frame_indices)
    # Ensure compatibility: Convert to NumPy array if necessary
    if hasattr(frames, "asnumpy"):  # It's a Tensor
        frames = frames.asnumpy()
    elif hasattr(frames, "numpy"):  # Older decord versions
        frames = frames.numpy()
    frames = [Image.fromarray(f).convert('RGB') for f in frames]
    return frames

def sample_gif(
        gif_path: str,
        n_frames:int = None,
        start_time: int = 0,
        end_time: int = -1
    ) -> List[Image.Image]:

    assert os.path.exists(gif_path), f"File not found: {gif_path}"
    
    gif_frames = Image.open(gif_path)

    start_frame = 0
    end_frame = gif_frames.n_frames - 1
    frame_indices = sample_frame_indices(
        start_frame=start_frame,
        total_frames=end_frame - start_frame + 1,
        n_frames=n_frames,
    )
        
    frames = []
    i = 0
    for frame in ImageSequence.Iterator(gif_frames):
        if i in frame_indices:
            frames.append(frame.convert('RGB'))
        i += 1
    return frames

def sample_image(
    image_path: str, 
    n_frames: int = None,
    start_time: int = 0,
    end_time: int = -1
    ):
    assert os.path.exists(image_path), f"File not found: {image_path}"
    image = Image.open(image_path).convert('RGB')
    return [image]

def get_visual_type(input_file):
    ext = os.path.splitext(input_file)[-1]
    if ext in {'.gif'}:
        return 'gif'
    elif ext in {'.mp4', '.avi', '.webm', '.mov', '.mkv', '.wmv'}:
        return 'video'
    elif ext in {'.jpg', '.jpeg', '.png', '.tif'}:
        return 'image'
    else:
        print(f"{VALID_DATA_FORMAT_STRING} But found {ext}!")
        return 'unk'

def get_benchmarks(benchmarks):
    final_benchmarks = []
    type2bm = {
        'dream': ['dream'],
        'caption': ['msvd-caption', 'msr-vtt-caption', 'vatex-caption'],
        'mc_qa': ['next-qa', 'egoschema', 'mvbench', 'video-mme'],
        'oe_qa': ['msvd-qa', 'msr-vtt-qa', 'tgif-qa', 'anet-qa'],
    }
    for bm in benchmarks:
        bm = bm.lower()
        if bm in final_benchmarks:
            continue
        if bm == 'all':
            for v in type2bm.values():
                final_benchmarks.extend(v)
            return final_benchmarks
        if bm in type2bm:
            final_benchmarks.extend(type2bm[bm])
        else:
            final_benchmarks.append(bm)
    return final_benchmarks

def check_data_format(data):
    for msg in data['messages']:
        if isinstance(msg['content'], dict):
            msg['content'] = [msg['content']]
        for content in msg['content']:
            assert content['type'] in {'image', 'video', 'text'}, f"content['type']={content['type']} MUST be one of ['image', 'video', 'text']"
            if content['type'] != "text":
                media_path_key = f"{content['type']}_file"
                meida_paths = content[content['type']][media_path_key]
                if isinstance(meida_paths, str):
                    meida_paths = [meida_paths]
                for path in meida_paths:
                    assert os.path.exists(path), f"File not found: {path}"

def format_one_sample(media_file=None, prompt="Describe the video in detail."):
    sample = {
        "messages": []
    }
    user_content = {
        "role": "user",
        "content": []
    }
    if media_file is not None:
        media_type = get_visual_type(media_file)
        if media_type in ("video", "gif"):
            media_type = "video"
        media_path_key = f"{media_type}_file"
        user_content["content"].append({
            "type": media_type,
            media_type: {
                media_path_key: media_file,
            }
        })
    user_content["content"].append({
        "type": "text",
        "text": prompt
    })

    assistant_content = {
        "role": "assistant",
        "content": []
    }

    sample["messages"].append(user_content)
    sample["messages"].append(assistant_content)
    if media_file is not None:
        sample["task"] = f"{media_type}/QA"
    else:
        sample["task"] = 'text-only'
    # check_data_format(sample)
    return sample


class DictToObject(object):
    def __init__(self, dictionary):
        for key, value in dictionary.items():
            setattr(self, key, value)


"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import torch
from ...common.registry import registry

from .blip import BlipBase
from .blip_outputs import (
    BlipOutput,
    BlipIntermediateOutput,
)
from ..med import XBertLMHeadDecoder
from ..vit import VisionTransformerEncoder


@registry.register_model("blip_caption")
class BlipCaption(BlipBase):
    """
    BLIP captioning model.

    Supported model types:
        - base_coco: fine-tuned BLIP base model on COCO caption dataset (Karparthy split).
        - large_coco: fine-tuned BLIP large model on COCO caption dataset (Karparthy split).

    Usage:
        >>> from lavis.models import load_model
        >>> model = load_model("blip_caption", "base_coco")
        >>> model = load_model("blip_caption", "large_coco")
    """

    PRETRAINED_MODEL_CONFIG_DICT = {
        "base_coco": "configs/models/blip_caption_base_coco.yaml",
        "large_coco": "configs/models/blip_caption_large_coco.yaml",
    }

    def __init__(self, image_encoder, text_decoder, prompt=None, max_txt_len=40):
        super().__init__()

        self.tokenizer = self.init_tokenizer()

        self.visual_encoder = image_encoder
        self.text_decoder = text_decoder

        self.prompt = prompt
        self.prompt_length = len(self.tokenizer(self.prompt).input_ids) - 1

        self.max_txt_len = max_txt_len

    def forward_encoder(self, samples):
        image_embeds = self.visual_encoder.forward_features(samples["image"])
        return image_embeds

    def forward_decoder(self, samples, image_embeds):
        # prepare inputs for forwarding decoder
        raw_text = samples["text_input"]
        text = self.tokenizer(
            raw_text,
            padding="longest",
            truncation=True,
            max_length=self.max_txt_len,
            return_tensors="pt",
        ).to(self.device)
        text.input_ids[:, 0] = self.tokenizer.bos_token_id

        # prepare targets for forwarding decoder
        decoder_targets = text.input_ids.masked_fill(
            text.input_ids == self.tokenizer.pad_token_id, -100
        )
        decoder_targets[:, : self.prompt_length] = -100

        # forward decoder
        image_atts = torch.ones(image_embeds.size()[:-1], dtype=torch.long).to(
            self.device
        )
        decoder_output = self.text_decoder(
            input_ids=text.input_ids,
            attention_mask=text.attention_mask,
            encoder_hidden_states=image_embeds,
            encoder_attention_mask=image_atts,
            labels=decoder_targets,
            return_dict=True,
        )

        return decoder_output, decoder_targets

    def forward(self, samples):
        r"""
        Args:
            samples (dict): A dictionary containing the following keys:
                - image (torch.Tensor): A tensor of shape (batch_size, 3, H, W)
                - text_input (list): A list of strings of length batch_size.
        Returns:
            output (BlipOutput): A BlipOutput object containing the following
                attributes:
                - loss (torch.Tensor): A scalar tensor containing the total loss. For BlipCaption, this is the same as the LM loss.
                - loss_lm (torch.Tensor): A scalar tensor containing the LM loss.
                - intermediate_outputs (BlipIntermediateOutput): A BlipIntermediateOutput object containing intermediate outputs.
                  see :class:`lavis.models.blip_models.blip_outputs.BlipOutput` for more details.

        Example:
        ```python
        >>> from PIL import Image
        >>> from lavis.models import load_model_and_preprocess
        >>> model, vis_processors, txt_processors = load_model_and_preprocess("blip_caption")
        >>> raw_image = Image.open("docs/data/merlion.png").convert("RGB")
        >>> image = vis_processors["eval"](raw_image).unsqueeze(0)
        >>> text_input = ["a large statue of a person spraying water from a fountain"]
        >>> samples = {"image": image, "text_input": text_input}
        >>> output = model(samples)
        >>> output.keys()
        odict_keys(['intermediate_output', 'loss', 'loss_lm'])
        >>> output.intermediate_output.image_embeds.shape
        torch.Size([1, 577, 768])
        >>> output.intermediate_output.decoder_labels.shape
        torch.Size([1, 13])
        ```"""

        image_embeds = self.forward_encoder(samples)
        decoder_output, decoder_targets = self.forward_decoder(samples, image_embeds)

        # return decoder_out
        return BlipOutput(
            loss=decoder_output.loss,
            loss_lm=decoder_output.loss,
            intermediate_output=BlipIntermediateOutput(
                image_embeds=image_embeds,
                decoder_output=decoder_output,
                decoder_labels=decoder_targets,
            ),
        )

    def generate(
        self,
        samples,
        use_nucleus_sampling=False,
        num_beams=3,
        max_length=30,
        min_length=10,
        top_p=0.9,
        repetition_penalty=1.0,
        num_captions=1,
    ):
        """
        Args:
            samples (dict): A dictionary containing the following keys:
                - image (torch.Tensor): A tensor of shape (batch_size, 3, H, W)
            use_nucleus_sampling (bool): Whether to use nucleus sampling. If False, use top-k sampling.
            num_beams (int): Number of beams for beam search. 1 means no beam search.
            max_length (int): The maximum length of the sequence to be generated.
            min_length (int): The minimum length of the sequence to be generated.
            top_p (float): The cumulative probability for nucleus sampling.
            repetition_penalty (float): The parameter for repetition penalty. 1.0 means no penalty.
            num_captions (int): Number of captions to be generated for each image.
        Returns:
            captions (list): A list of strings of length batch_size * num_captions.

        Example:
        ```python
        >>> from PIL import Image
        >>> from lavis.models import load_model_and_preprocess
        >>> model, vis_processors, txt_processors = load_model_and_preprocess("blip_caption")
        >>> raw_image = Image.open("docs/data/merlion.png").convert("RGB")
        >>> image = vis_processors["eval"](raw_image).unsqueeze(0)
        >>> samples = {"image": image}
        >>> captions = model.generate(samples)
        >>> captions
        ['a large statue of a person spraying water from a fountain']
        >>> captions = model.generate(samples, use_nucleus_sampling=True, num_captions=3)
        >>> captions # example output, results may vary due to randomness
        ['singapore showing the view of some building',
        'the singapore harbor in twilight, as the weather is going down',
        'the famous singapore fountain at sunset']
        """
        # prepare inputs for decoder generation.
        encoder_out = self.forward_encoder(samples)
        image_embeds = torch.repeat_interleave(encoder_out, num_captions, 0)

        prompt = [self.prompt] * image_embeds.size(0)
        prompt = self.tokenizer(prompt, return_tensors="pt").to(self.device)
        prompt.input_ids[:, 0] = self.tokenizer.bos_token_id
        prompt.input_ids = prompt.input_ids[:, :-1]

        # get decoded text
        decoder_out = self.text_decoder.generate_from_encoder(
            tokenized_prompt=prompt,
            visual_embeds=image_embeds,
            sep_token_id=self.tokenizer.sep_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
            use_nucleus_sampling=use_nucleus_sampling,
            num_beams=num_beams,
            max_length=max_length,
            min_length=min_length,
            top_p=top_p,
            repetition_penalty=repetition_penalty,
        )

        outputs = self.tokenizer.batch_decode(decoder_out, skip_special_tokens=True)
        captions = [output[len(self.prompt) :] for output in outputs]

        return captions

    @classmethod
    def from_config(cls, cfg):
        # vision encoder
        image_encoder = VisionTransformerEncoder.from_config(cfg)
        # text encoder + multimodal decoder
        text_decoder = XBertLMHeadDecoder.from_config(cfg)

        prompt = cfg.get("prompt", None)
        max_txt_len = cfg.get("max_txt_len", 40)

        model = cls(image_encoder, text_decoder, prompt=prompt, max_txt_len=max_txt_len)
        model.load_checkpoint_from_config(cfg)

        return model



"""
Ayase - Modular media quality metrics toolkit

"""

__version__ = "0.1.23"
__author__ = "Ayase Contributors"
__description__ = "Modular media quality metrics toolkit"

from .pipeline import AyasePipeline
from .profile import PipelineProfile, instantiate_profile_modules, load_profile

__all__ = [
    "__version__",
    "__author__",
    "__description__",
    "AyasePipeline",
    "PipelineProfile",
    "load_profile",
    "instantiate_profile_modules",
]



"""FAD — Frechet Audio Distance (2019).

Dataset-level metric that measures the distance between distributions of
audio features, analogous to FID for images. Uses VGGish or similar audio
embeddings and computes the Frechet distance.

pip install frechet_audio_distance

fad_score — lower = better (closer audio distributions).
"""

import logging
import subprocess
import tempfile
from pathlib import Path
from typing import Optional, List

import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import BatchMetricModule

logger = logging.getLogger(__name__)


class FADModule(BatchMetricModule):
    name = "fad"
    description = "Frechet Audio Distance for audio generation (batch metric, 2019)"
    default_config = {
        "subsample_videos": None,
        "sample_rate": 16000,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._ml_available = False
        self._backend = None
        self.subsample_videos = self.config.get("subsample_videos", None)
        self.sample_rate = self.config.get("sample_rate", 16000)
        self._processed_count = 0

    def setup(self) -> None:
        # Tier 1: frechet_audio_distance package
        try:
            from frechet_audio_distance import FrechetAudioDistance
            self._model = FrechetAudioDistance()
            self._ml_available = True
            self._backend = "fad_package"
            logger.info("FAD module initialised (frechet_audio_distance package)")
            return
        except ImportError:
            pass
        except Exception as e:
            logger.debug(f"FAD package init failed: {e}")

        # Tier 2: heuristic (spectral features + Frechet distance)
        logger.info("FAD module initialised (heuristic fallback)")

    def extract_features(self, sample: Sample) -> Optional[np.ndarray]:
        """Extract audio features for distribution comparison."""
        if self.subsample_videos is not None and self._processed_count >= self.subsample_videos:
            return None

        try:
            audio = self._load_audio(sample.path)
            if audio is None:
                return None

            features = self._compute_spectral_features(audio)
            if features is not None:
                self._processed_count += 1
            return features
        except Exception as e:
            logger.debug(f"FAD feature extraction failed for {sample.path}: {e}")
            return None

    def _load_audio(self, path: Path) -> Optional[np.ndarray]:
        """Load audio from file, extracting from video if necessary."""
        # Try direct audio loading
        try:
            import soundfile as sf
            audio, sr = sf.read(str(path))
            if audio.ndim > 1:
                audio = audio.mean(axis=1)
            if sr != self.sample_rate:
                # Simple resampling via linear interpolation
                duration = len(audio) / sr
                n_samples = int(duration * self.sample_rate)
                indices = np.linspace(0, len(audio) - 1, n_samples)
                audio = np.interp(indices, np.arange(len(audio)), audio)
            return audio.astype(np.float32)
        except Exception:
            pass

        # Try extracting audio from video via ffmpeg
        try:
            tmp = tempfile.NamedTemporaryFile(suffix=".wav", delete=False)
            tmp.close()
            cmd = [
                "ffmpeg", "-y", "-i", str(path),
                "-vn", "-ac", "1", "-ar", str(self.sample_rate),
                "-sample_fmt", "s16", tmp.name,
            ]
            result = subprocess.run(cmd, capture_output=True, timeout=30)
            if result.returncode != 0:
                Path(tmp.name).unlink(missing_ok=True)
                return None

            import soundfile as sf
            audio, _ = sf.read(tmp.name, dtype="float32")
            Path(tmp.name).unlink(missing_ok=True)
            if audio.ndim > 1:
                audio = audio.mean(axis=1)
            return audio
        except Exception:
            pass

        return None

    def _compute_spectral_features(self, audio: np.ndarray) -> Optional[np.ndarray]:
        """Compute spectral features (MFCCs + spectral statistics)."""
        if len(audio) < 1024:
            return None

        # Compute spectrogram
        n_fft = 1024
        hop = 512
        n_frames = (len(audio) - n_fft) // hop + 1
        if n_frames < 1:
            return None

        frames = np.stack([
            audio[i * hop:i * hop + n_fft] * np.hanning(n_fft)
            for i in range(n_frames)
        ])
        spectrogram = np.abs(np.fft.rfft(frames, axis=1))

        # Mel-like features: group FFT bins into ~20 bands
        n_bands = 20
        n_bins = spectrogram.shape[1]
        band_size = max(n_bins // n_bands, 1)
        mel_features = []
        for b in range(n_bands):
            start = b * band_size
            end = min(start + band_size, n_bins)
            if start >= n_bins:
                mel_features.append(0.0)
            else:
                mel_features.append(float(spectrogram[:, start:end].mean()))

        # Spectral statistics
        spectral_centroid = float(np.mean(
            np.sum(spectrogram * np.arange(spectrogram.shape[1]), axis=1) /
            (np.sum(spectrogram, axis=1) + 1e-8)
        ))
        spectral_rolloff = float(np.mean(np.percentile(spectrogram, 85, axis=1)))
        spectral_flux = float(np.mean(np.diff(spectrogram, axis=0) ** 2)) if n_frames > 1 else 0.0

        features = np.array(mel_features + [spectral_centroid, spectral_rolloff, spectral_flux])
        return features.astype(np.float64)

    def compute_distribution_metric(
        self, features: List[np.ndarray], reference_features: Optional[List[np.ndarray]] = None
    ) -> float:
        """Compute Frechet distance between audio feature distributions."""
        try:
            features_array = np.stack(features, axis=0)

            if reference_features is not None and len(reference_features) > 0:
                ref_array = np.stack(reference_features, axis=0)
            else:
                mid = len(features_array) // 2
                if mid < 1:
                    return 0.0
                ref_array = features_array[:mid]
                features_array = features_array[mid:]

            return self._frechet_distance(features_array, ref_array)
        except Exception as e:
            logger.error(f"FAD computation failed: {e}")
            return float("inf")

    def _frechet_distance(self, feat1: np.ndarray, feat2: np.ndarray) -> float:
        mu1 = np.mean(feat1, axis=0)
        mu2 = np.mean(feat2, axis=0)

        if feat1.shape[0] < 2 or feat2.shape[0] < 2:
            return float(np.sum((mu1 - mu2) ** 2))

        sigma1 = np.cov(feat1, rowvar=False)
        sigma2 = np.cov(feat2, rowvar=False)

        if sigma1.ndim == 0:
            sigma1 = np.array([[sigma1]])
        if sigma2.ndim == 0:
            sigma2 = np.array([[sigma2]])

        diff = mu1 - mu2

        try:
            from scipy import linalg
            covmean, _ = linalg.sqrtm(sigma1 @ sigma2, disp=False)
            if np.iscomplexobj(covmean):
                covmean = covmean.real
            fd = diff @ diff + np.trace(sigma1 + sigma2 - 2 * covmean)
        except ImportError:
            fd = float(diff @ diff + np.trace(sigma1) + np.trace(sigma2))

        return float(fd)

    def on_dispose(self) -> None:
        if len(self._feature_cache) < 2:
            logger.info(f"FAD: Not enough samples ({len(self._feature_cache)})")
            self._feature_cache = []
            self._reference_cache = []
            return

        try:
            score = self.compute_distribution_metric(
                self._feature_cache,
                self._reference_cache if self._reference_cache else None,
            )
            logger.info(f"FAD: {score:.4f} ({len(self._feature_cache)} samples)")

            if hasattr(self, "pipeline") and self.pipeline:
                if hasattr(self.pipeline, "add_dataset_metric"):
                    self.pipeline.add_dataset_metric("fad", score)
        except Exception as e:
            logger.error(f"FAD failed: {e}")
        finally:
            self._feature_cache = []
            self._reference_cache = []
            self._processed_count = 0



import torch
import numpy as np
from PIL import Image
from typing import List, Union
from transformers import AutoModelForCausalLM, AutoTokenizer, AutoProcessor
from decord import VideoReader, cpu

from .vqa_model import VQAScoreModel

MPLUG_OWL3_MODELS = {
    'mplug-owl3-7b': {
        'tokenizer': {
            'path': 'mPLUG/mPLUG-Owl3-7B-240728',
        },
        'model': {
            'pretrained_model_name_or_path': 'mPLUG/mPLUG-Owl3-7B-240728',
            'attn_implementation': 'flash_attention_2',
            'trust_remote_code': True,
            'torch_dtype': torch.bfloat16,
        },
    },
}

class mPLUGOwl3Model(VQAScoreModel):
    video_mode = "direct"
    allows_image = True
    def __init__(self,
                 model_name='mplug-owl3-7b',
                 device='cuda',
                 cache_dir=None):
        assert model_name in MPLUG_OWL3_MODELS, f"Model {model_name} not found in MPLUG_OWL3_MODELS"
        self.model_name = model_name
        self.device = device
        self.cache_dir = cache_dir
        self.model_info = MPLUG_OWL3_MODELS[model_name]
        self.load_model()
        

    def load_model(self):
        # model_path = self.model_info['model']['path']
        tokenizer_path = self.model_info['tokenizer']['path']
        
        self.model = AutoModelForCausalLM.from_pretrained(
            # model_path, 
            **self.model_info['model']
        ).to(self.device)
        
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
        self.processor = self.model.init_processor(self.tokenizer)

    def load_images(self, paths: List[str], num_frames: int = 16) -> List[Union[torch.Tensor, List[Image.Image]]]:
        processed_data = []
        for path in paths:
            if path.lower().endswith(('.mp4', '.avi', '.mov', '.mkv')):  # Video file
                video_frames = self.encode_video(path, num_frames)
                processed_data.append(video_frames)
            elif path.lower().endswith('.npy'):  # NumPy file
                np_array = np.load(path)
                if np_array.ndim == 3:  # Single image
                    image = Image.fromarray(np_array.astype('uint8'), 'RGB')
                    processed_data.append(image)
                elif np_array.ndim == 4:  # Multiple frames
                    frames = [Image.fromarray(frame.astype('uint8'), 'RGB') for frame in np_array]
                    processed_data.append(frames)
                else:
                    raise ValueError(f"Unexpected shape for NumPy array in {path}")
            else:  # Regular image file
                image = Image.open(path).convert('RGB')
                processed_data.append(image)
        return processed_data

    def encode_video(self, video_path, max_frames_num):
        def uniform_sample(l, n):
            gap = len(l) / n
            idxs = [int(i * gap + gap / 2) for i in range(n)]
            return [l[i] for i in idxs]

        vr = VideoReader(video_path, ctx=cpu(0))
        sample_fps = round(vr.get_avg_fps() / 1)  # FPS
        frame_idx = [i for i in range(0, len(vr), sample_fps)]
        if len(frame_idx) > max_frames_num:
            frame_idx = uniform_sample(frame_idx, max_frames_num)
        frames = vr.get_batch(frame_idx).numpy()
        frames = [Image.fromarray(v.astype('uint8')) for v in frames]
        return frames

    def forward(self,
                paths: List[str],
                texts: List[str],
                num_frames: int=16,
                question_template: str = "Does this figure show \"{}\"? Please answer yes or no.",
                answer_template: str = "Yes") -> torch.Tensor:
        assert len(paths) == len(texts), "Number of paths and texts must match"

        questions = [question_template.format(text) for text in texts]
        answers = [answer_template.format(text) for text in texts]
        processed_data = self.load_images(paths, num_frames)

        lm_probs = []
        for data, question, answer in zip(processed_data, questions, answers):
            if isinstance(data, list):  # Video
                messages = [
                    {"role": "user", "content": f"<|video|>\n{question}"},
                    {"role": "assistant", "content": ""}
                ]
                inputs = self.processor(messages, images=None, videos=[data])
            else:  # Image
                messages = [
                    {"role": "user", "content": f"<|image|>\n{question}"},
                    {"role": "assistant", "content": ""}
                ]
                inputs = self.processor(messages, images=[data], videos=None)

            inputs.to(self.device)

            with torch.inference_mode():
                image_embeds = self.model.forward_image(inputs['pixel_values'])

                terminators = [self.tokenizer.convert_tokens_to_ids(i) for i in self.model.terminators]

                outputs = self.model.language_model.generate(
                    input_ids=inputs['input_ids'],
                    image_embeds=image_embeds,
                    media_offset=inputs['media_offset'],
                    pad_token_id=0,
                    eos_token_id=terminators,
                    attention_mask=None,
                    max_new_tokens=1,
                    do_sample=False,
                    output_scores=True,
                    return_dict_in_generate=True
                )
            
            scores = outputs.scores[0]
            probs = torch.nn.functional.softmax(scores, dim=-1)
            yes_token_id = self.tokenizer.encode(answer)[0]
            lm_prob = probs[0, yes_token_id].item()
            lm_probs.append(lm_prob)

        return torch.tensor(lm_probs)
    

    def generate(self,
                images: List[str],
                texts: List[str],
                num_frames: int=16,
                max_new_tokens: int = 256) -> List[str]:
        assert len(images) == len(texts), "Number of paths and texts must match"

        processed_data = self.load_images(images, num_frames)

        generated_texts = []
        for data, text in zip(processed_data, texts):
            if isinstance(data, list):  # Video
                messages = [
                    {"role": "user", "content": f"<|video|>\n{text}"},
                    {"role": "assistant", "content": ""}
                ]
                inputs = self.processor(messages, images=None, videos=[data])
            else:  # Image
                messages = [
                    {"role": "user", "content": f"<|image|>\n{text}"},
                    {"role": "assistant", "content": ""}
                ]
                inputs = self.processor(messages, images=[data], videos=None)

            inputs.to(self.device)

            with torch.inference_mode():
                image_embeds = self.model.forward_image(inputs['pixel_values'])

                terminators = [self.tokenizer.convert_tokens_to_ids(i) for i in self.model.terminators]

                outputs = self.model.language_model.generate(
                    input_ids=inputs['input_ids'],
                    image_embeds=image_embeds,
                    media_offset=inputs['media_offset'],
                    pad_token_id=0,
                    eos_token_id=terminators,
                    attention_mask=None,
                    max_new_tokens=max_new_tokens,
                    do_sample=False
                )
                
                text = self.tokenizer.decode(outputs[0], skip_special_tokens=True).split('\nassistant\n')[1]
                generated_texts.append(text.strip())

        return generated_texts